# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import Optional, cast, Union, TYPE_CHECKING, List, Tuple, Generator, Any

from qt import Qt, QPointF, QSize, QModelIndex, QEvent, QTimer
from qt import QColor, QPushButton, QPixmapCache
from qt import QWidget, QLabel, QTabWidget
from qt import QGridLayout, QVBoxLayout, QHBoxLayout, QSpinBox
from qt import QDialog, QStringListModel, QListView, QSplitter, QValidator
from qt import QIcon, QPixmap, QPainter, QDialogButtonBox
from qt import QSizePolicy, QComboBox, QCheckBox, QScrollBar
from qt import QAbstractItemView, QHeaderView
from qt import QTreeView, QFont, QFrame
from qt import QStyledItemDelegate
from qt import QBrush, QPalette
from qt import PYQT6
from kde import KDialogButtonBox, KApplication

from modeltest import ModelTest

from rulesetselector import RuleTreeView
from board import WindLabel
from mi18n import i18n, i18nc
from common import Internal, Debug
from statesaver import StateSaver
from query import Query
from guiutil import ListComboBox, Painter, decorateWindow, BlockSignals
from tree import TreeItem, RootItem, TreeModel
from tile import Tile, MeldList
from point import Point

if TYPE_CHECKING:
    from qt import QObject, QRect, QStyleOptionViewItem, QPersistentModelIndex
    from tile import Meld
    from rule import Rule, Score
    from scene import ScoringScene
    from uitile import UITile
    from hand import Hand
    from scoring import ScoringGame, ScoringPlayer
    from player import Player

class ScoreTreeItem(TreeItem):

    """generic class for items in our score tree"""
    # pylint: disable=abstract-method
    # we know content() is abstract, this class is too

    def columnCount(self) ->int:
        """count the hands of the first player"""
        child1 = self
        while not isinstance(child1, ScorePlayerItem) and child1.children:
            child1 = cast('ScoreTreeItem', child1.children[0])
        if isinstance(child1, ScorePlayerItem):
            return len(child1.raw[1]) + 1
        return 1


class ScoreRootItem(RootItem):

    """the root item for the score tree"""

    def columnCount(self) ->int:
        child1 = self
        while not isinstance(child1, ScorePlayerItem) and child1.children:
            child1 = cast('ScoreRootItem', child1.children[0])
        if isinstance(child1, ScorePlayerItem):
            return len(child1.raw[1]) + 1
        return 1


class ScoreGroupItem(ScoreTreeItem):

    """represents a group in the tree like Points, Payments, Balance"""

    def __init__(self, content:str) ->None:
        super().__init__(content)

    def content(self, column:int) ->str:
        """return content stored in this item"""
        return i18n(self.raw)


class ScorePlayerItem(ScoreTreeItem):

    """represents a player in the tree"""

    def __init__(self, content:Tuple[str, List[Any]]) ->None:
        super().__init__(content)

    def content(self, column:int) ->Union[str, Any, None]:
        """return the content stored in this node"""
        if column == 0:
            return i18n(self.raw[0])
        try:
            return self.hands()[column - 1]
        except IndexError:
            # we have a penalty but no hand yet. Should
            # not happen in practical use
            return None

    def hands(self) ->List[Any]:
        """a small helper"""
        return self.raw[1]

    def chartPoints(self, column:int, steps:int) ->Generator[float, None, None]:
        """the returned points spread over a height of four rows"""
        int_points = [x.balance for x in self.hands()]
        int_points.insert(0, 0)
        int_points.insert(0, 0)
        int_points.append(int_points[-1])
        column -= 1
        int_points = int_points[column:column + 4]
        points = [float(x) for x in int_points]
        for idx in range(1, len(points) - 2):  # skip the ends
            for step in range(steps):
                point_1, point0, point1, point2 = points[idx - 1:idx + 3]
                fstep = float(step) / steps
                # wikipedia Catmull-Rom -> Cubic_Hermite_spline
                # 0 -> point0, 1 -> point1, 1/2 -> (- point_1 + 9 point0 + 9
                # point1 - point2) / 16
                yield (
                    fstep * ((2 - fstep) * fstep - 1) * point_1
                    + (fstep * fstep * (
                        3 * fstep - 5) + 2) * point0
                    + fstep *
                    ((4 - 3 * fstep) * fstep + 1) * point1
                    + (fstep - 1) * fstep * fstep * point2) / 2
        yield points[-2]


class ScoreItemDelegate(QStyledItemDelegate):

    """since setting delegates for a row does not work as wanted with a
    tree view, we set the same delegate on ALL items."""
    # try to use colors that look good with all color schemes. Bright
    # contrast colors are not optimal as long as our lines have a width of
    # only one pixel: antialiasing is not sufficient
    colors = [KApplication.palette().color(x)
              for x in [QPalette.ColorRole.Text, QPalette.ColorRole.Link, QPalette.ColorRole.LinkVisited]]
    colors.append(QColor('orange'))

    def __init__(self, parent:Optional['QObject']=None) ->None:
        super().__init__(parent)

    def paint(self, painter:Optional[QPainter], option:'QStyleOptionViewItem',
        index:Union[QModelIndex,'QPersistentModelIndex']) ->None:
        """where the real work is done..."""
        assert painter
        assert isinstance(index, QModelIndex), index
        item = index.internalPointer()
        if isinstance(item, ScorePlayerItem) and item.parent and item.parent.row() == 3 and index.column() != 0:
            parent_item = cast(TreeItem, index.parent().internalPointer())
            for idx, playerItem in enumerate(parent_item.children):
                assert isinstance(playerItem, ScorePlayerItem), playerItem
                rect = option.rect  # type:ignore[attr-defined]
                chart = cast('ScoreModel', index.model()).chart(rect, index, playerItem)
                if chart:
                    with Painter(painter):
                        painter.translate(rect.topLeft())
                        painter.setPen(self.colors[idx])
                        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
                        # if we want to use a pen width > 1, we can no longer directly drawPolyline
                        # separately per cell beause the lines spread vertically over two rows: We would
                        # have to draw the lines into one big pixmap and copy
                        # from the into the cells
                        if PYQT6:
                            # seems the annotations are wrong: list(QPointF) does work
                            painter.drawPolyline(chart)  # type:ignore[call-overload]
                        else:
                            painter.drawPolyline(chart)
            return
        super().paint(painter, option, index)


class ScoreModel(TreeModel):

    """a model for our score table"""
    steps = 30  # how fine do we want the stepping in the chart spline

    tupleType: Any = type(None)

    def __init__(self, scoreTable:'ScoreTable', parent:Optional['QObject']=None) ->None:
        super().__init__(parent)
        self.scoreTable = scoreTable
        self.rootItem = ScoreRootItem(None)
        self.minY = 9999999.9
        self.maxY = -9999999.9
        self.loadData()

    def chart(self, rect:'QRect', index:QModelIndex, playerItem:'ScorePlayerItem') ->List[QPointF]:
        """return list(QPointF) for a player in a specific tree cell"""
        chartHeight = float(rect.height()) * 4
        yScale = chartHeight / (self.minY - self.maxY)
        yOffset = rect.height() * index.row()
        _ = (playerItem.chartPoints(index.column(), self.steps))
        yValues = [(y - self.maxY) * yScale - yOffset for y in _]
        stepX = float(rect.width()) / self.steps
        xValues = [x * stepX for x in range(self.steps + 1)]
        return [QPointF(x, y) for x, y in zip(xValues, yValues)]

    def data(self, index:Union[QModelIndex,'QPersistentModelIndex'], role:int=Qt.ItemDataRole.DisplayRole) ->Any:  # pylint: disable=too-many-branches
        """score table"""
        # pylint: disable=too-many-return-statements
        assert isinstance(index, QModelIndex)
        if not index.isValid():
            return None
        column = index.column()
        item = cast(TreeItem, index.internalPointer())
        assert item.parent
        if role == Qt.ItemDataRole.DisplayRole:
            if isinstance(item, ScorePlayerItem):
                content = item.content(column)
                # if content.__class__.__name__ == 'Score':
                if isinstance(content, ScoreModel.tupleType):
                    parentRow = item.parent.row()
                    if parentRow == 0:
                        if not content.penalty:
                            content = f'{int(content.points)} {content.wind}'
                    elif parentRow == 1:
                        content = str(content.payments)
                    else:
                        content = str(content.balance)
                return content
            return '' if column > 0 else item.content(0)
        if role == Qt.ItemDataRole.TextAlignmentRole:
            return int(Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter) \
                if index.column() == 0 else int(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        if role == Qt.ItemDataRole.FontRole:
            return QFont('Monospaced')
        if role == Qt.ItemDataRole.ForegroundRole:
            if isinstance(item, ScorePlayerItem) and item.parent.row() == 3:
                content = item.content(column)
                if not isinstance(content, ScoreModel.tupleType):
                    return QBrush(ScoreItemDelegate.colors[index.row()])
        if column > 0 and isinstance(item, ScorePlayerItem):
            content = item.content(column)  # type:ignore
            if role == Qt.ItemDataRole.BackgroundRole:
                assert isinstance(content, ScoreModel.tupleType)
                if content and content.won:
                    return QColor(165, 255, 165)
            if role == Qt.ItemDataRole.ToolTipRole:
                englishHints = content.manualrules.split('||')  # type:ignore
                tooltip = '<br />'.join(i18n(x) for x in englishHints)
                return tooltip
        return None

    def headerData(self, section:int, orientation:Qt.Orientation, role:int=Qt.ItemDataRole.DisplayRole) ->Any:
        """tell the view about the wanted headers"""
        if role == Qt.ItemDataRole.DisplayRole and orientation == Qt.Orientation.Horizontal:
            if section == 0:
                return i18n('Round/Hand')
            assert self.rootItem
            child1 = self.rootItem.children[0]
            if child1 and child1.children:
                child1 = cast(ScorePlayerItem, child1.children[0])
                hands = child1.hands()
                handResult = hands[section - 1]
                if not handResult.penalty:
                    return self.handTitle(handResult)
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            return int(Qt.AlignmentFlag.AlignLeft | Qt.AlignmentFlag.AlignVCenter) \
                if section == 0 else int(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
        return None

    def loadData(self) ->None:
        """loads all data from the data base into a 2D matrix formatted like the wanted tree"""
        game = self.scoreTable.game
        assert game
        assert game.gameid

        fields = 'game,hand,player,rotated,notrotated,penalty,won,prevailing,wind,points,payments,balance,manualrules'
        tuples = Query('select {fields} from score where game=? order by hand', (game.gameid, ),
            fields=fields).tuples()
#        if not tuples:
#            return
        if tuples:
            ScoreModel.tupleType = tuples[0].__class__
        humans = sorted(
            (x for x in game.players if not x.name.startswith('Robot')))
        robots = sorted(
            (x for x in game.players if x.name.startswith('Robot')))
        data =  cast(List[Tuple[str, List[Any]]],
                    [tuple([player.localName, [x for x in tuples  # type:ignore
                    if x.player == player.nameid]]) for player in humans + robots])
#        print(f'players: {Players.allNames}')
#        for idx, _ in enumerate(data):
#            print(f'{idx}: {_}')
        self.__findMinMaxChartPoints(data)
        parent = QModelIndex()
        assert self.rootItem
        groupIndex = self.index(self.rootItem.childCount(), 0, parent)
        groupNames = [i18nc('kajongg', 'Score'), i18nc('kajongg', 'Payments'),
                      i18nc('kajongg', 'Balance'), i18nc('kajongg', 'Chart')]
        for idx, groupName in enumerate(groupNames):
            self.insertRows(idx, list([ScoreGroupItem(groupName)]), groupIndex)
            listIndex = self.index(idx, 0, groupIndex)
            for idx1, item in enumerate(data):
                self.insertRows(idx1, list([ScorePlayerItem(item)]), listIndex)

    def __findMinMaxChartPoints(self, data:List[Tuple[str, List[Any]]]) ->None:
        """find and save the extremes of the spline. They can be higher than
        the pure balance values"""
        self.minY = 9999999.9
        self.maxY = -9999999.9
        for item in data:
            playerItem = ScorePlayerItem(item)
            for col in range(len(playerItem.hands())):
                points = list(playerItem.chartPoints(col + 1, self.steps))
                self.minY = min(self.minY, min(points))  # pylint:disable=nested-min-max
                self.maxY = max(self.maxY, max(points))  # pylint:disable=nested-min-max
        self.minY -= 2  # antialiasing might cross the cell border
        self.maxY += 2

    def handTitle(self, handResult:Any) ->str:
        """identifies the hand for window title and scoring table"""
        return str(Point(handResult))


class ScoreViewLeft(QTreeView):

    """subclass for defining sizeHint"""

    def __init__(self, parent:Optional[QWidget]=None) ->None:
        super().__init__(parent)
        self.setItemDelegate(ScoreItemDelegate(self))

    def __col0Width(self) ->int:
        """the width we need for displaying column 0
        without scrollbar"""
        return self.columnWidth(0) + self.frameWidth() * 2

    def sizeHint(self) ->QSize:
        """we never want a horizontal scrollbar for player names,
        we always want to see them in full"""
        return QSize(self.__col0Width(), QTreeView.sizeHint(self).height())

    def minimumSizeHint(self) ->QSize:
        """we never want a horizontal scrollbar for player names,
        we always want to see them in full"""
        return self.sizeHint()


class ScoreViewRight(QTreeView):

    """we need to subclass for catching events"""

    def __init__(self, parent:Optional[QWidget]=None) ->None:
        super().__init__(parent)
        self.setItemDelegate(ScoreItemDelegate(self))

    def changeEvent(self, event:Optional[QEvent]) ->None:
        """recompute column width if font changes"""
        if event:
            if event.type() == QEvent.Type.FontChange:
                self.setColWidth()

    def setColWidth(self) ->None:
        """we want a fixed column width sufficient for all values"""
        if header := self.header():
            colRange = range(1, header.count())
            if colRange:
                for col in colRange:
                    self.resizeColumnToContents(col)
                width = max(self.columnWidth(x) for x in colRange)
                for col in colRange:
                    self.setColumnWidth(col, width)


class HorizontalScrollBar(QScrollBar):

    """We subclass here because we want to react on show/hide"""

    def __init__(self, scoreTable:'ScoreTable', parent:Optional[QWidget]=None) ->None:
        super().__init__(parent)
        self.scoreTable = scoreTable

    def showEvent(self, unusedEvent:Optional[QEvent]) ->None:
        """adjust the left view"""
        self.scoreTable.adaptLeftViewHeight()

    def hideEvent(self, unusedEvent:Optional[QEvent]) ->None:
        """adjust the left view"""
        if header := self.scoreTable.viewRight.header():
            header.setOffset(0)  # we should not have to do this...
        # how to reproduce problem without setOffset:
        # show table with hor scroll, scroll to right, extend window
        # width very fast. The faster we do that, the wronger the
        # offset of the first column in the viewport.
        self.scoreTable.adaptLeftViewHeight()


class ScoreTable(QWidget):

    """show scores of current or last game, even if the last game is
    finished. To achieve this we keep our own reference to game."""

    def __init__(self, scene:'ScoringScene') ->None:
        super().__init__(None)
        self.setObjectName('ScoreTable')
        self.scene = scene
        self.scoreModel:Optional[ScoreModel] = None
        self.scoreModelTest:Optional[ModelTest] = None
        decorateWindow(self, i18nc('kajongg', 'Scores'))
        self.setAttribute(Qt.WidgetAttribute.WA_AlwaysShowToolTips)
        self.setWindowFlags(self.windowFlags() | Qt.WindowType.Tool)
        self.setMouseTracking(True)
        self.setupUi()
        self.refresh()
        StateSaver(self, self.splitter)

    @property
    def game(self) ->'ScoringGame':
        """a proxy"""
        return self.scene.game

    def setColWidth(self) ->None:
        """we want to accommodate for 5 digits plus minus sign
        and all column widths should be the same, making
        horizontal scrolling per item more pleasant"""
        self.viewRight.setColWidth()

    def setupUi(self) ->None:
        """setup UI elements"""
        self.viewLeft = ScoreViewLeft(self)
        self.viewRight = ScoreViewRight(self)
        self.viewRight.setHorizontalScrollBar(HorizontalScrollBar(self))
        self.viewRight.setHorizontalScrollMode(QAbstractItemView.ScrollMode.ScrollPerItem)
        self.viewRight.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        if header := self.viewRight.header():
            header.setSectionsClickable(False)
            header.setSectionsMovable(False)
        self.viewRight.setSelectionMode(QAbstractItemView.SelectionMode.NoSelection)
        windowLayout = QVBoxLayout(self)
        self.splitter = QSplitter(Qt.Orientation.Vertical)
        self.splitter.setObjectName('ScoreTableSplitter')
        windowLayout.addWidget(self.splitter)
        scoreWidget = QWidget()
        self.scoreLayout = QHBoxLayout(scoreWidget)
        leftLayout = QVBoxLayout()
        leftLayout.addWidget(self.viewLeft)
        self.leftLayout = leftLayout
        self.scoreLayout.addLayout(leftLayout)
        self.scoreLayout.addWidget(self.viewRight)
        self.splitter.addWidget(scoreWidget)
        self.ruleTree = RuleTreeView(i18nc('kajongg', 'Used Rules'))
        self.splitter.addWidget(self.ruleTree)
        # this shows just one line for the ruleTree - so we just see the
        # name of the ruleset:
        self.splitter.setSizes(list([1000, 1]))

    def sizeHint(self) ->QSize:
        """give the scoring table window a sensible default size"""
        result = QWidget.sizeHint(self)
        result.setWidth(result.height() * 3 // 2)
        # the default is too small. Use at least 2/5 of screen height and 1/4
        # of screen width:
        available = KApplication.desktopSize()
        height = max(result.height(), available.height() * 2 // 5)
        width = max(result.width(), available.width() // 4)
        result.setHeight(height)
        result.setWidth(width)
        return result

    def refresh(self) ->None:
        """load this game and this player. Keep parameter list identical with
        ExplainView"""
        # pylint:disable=too-many-branches
        if not self.game:
            # keep scores of previous game on display
            return
        if self.scoreModel:
            expandGroups = [
                self.viewLeft.isExpanded(
                    self.scoreModel.index(x, 0, QModelIndex()))
                for x in range(4)]
        else:
            expandGroups = [True, False, True, True]
        gameid = str(self.game.seed or self.game.gameid)
        if self.game.finished():
            title = i18n('Final scores for game <numid>%1</numid>', gameid)
        else:
            title = i18n('Scores for game <numid>%1</numid>', gameid)
        decorateWindow(self, title)
        self.ruleTree.rulesets = list([self.game.ruleset])
        self.scoreModel = ScoreModel(self)
        if Debug.modelTest:
            self.scoreModelTest = ModelTest(self.scoreModel, self)
        for view in [self.viewLeft, self.viewRight]:
            view.setModel(self.scoreModel)
            if header := view.header():
                header.setStretchLastSection(False)
            view.setAlternatingRowColors(True)
        left_header = self.viewLeft.header()
        right_header = self.viewRight.header()
        if not left_header or not right_header:
            return
        right_header.setSectionResizeMode(QHeaderView.ResizeMode.Fixed)
        for col in range(left_header.count()):
            left_header.setSectionHidden(col, col > 0)
            right_header.setSectionHidden(col, col == 0)
        self.scoreLayout.setStretch(1, 100)
        self.scoreLayout.setSpacing(0)
        self.viewLeft.setFrameStyle(QFrame.Shape.NoFrame)
        self.viewRight.setFrameStyle(QFrame.Shape.NoFrame)
        self.viewLeft.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        for master, slave in ((self.viewRight, self.viewLeft), (self.viewLeft, self.viewRight)):
            master.expanded.connect(slave.expand)
            master.collapsed.connect(slave.collapse)
            master_sb = master.verticalScrollBar()
            slave_sb = slave.verticalScrollBar()
            if master_sb and slave_sb:
                master_sb.valueChanged.connect(slave_sb.setValue)
        for row, expand in enumerate(expandGroups):
            self.viewLeft.setExpanded(
                self.scoreModel.index(row,
                                      0,
                                      QModelIndex()),
                expand)
        self.viewLeft.resizeColumnToContents(0)
        self.viewRight.setColWidth()
        # we need a timer since the scrollbar is not yet visible
        QTimer.singleShot(0, self.scrollRight)

    def scrollRight(self) ->None:
        """make sure the latest hand is visible"""
        if scrollBar := self.viewRight.horizontalScrollBar():
            scrollBar.setValue(scrollBar.maximum())

    def showEvent(self, unusedEvent:Optional[QEvent]) ->None:
        """Only now the views and scrollbars have useful sizes, so we can compute the spacer
        for the left view"""
        self.adaptLeftViewHeight()

    def adaptLeftViewHeight(self) ->None:
        """if the right view has a horizontal scrollbar, make sure both
        view have the same vertical scroll area. Otherwise scrolling to
        bottom results in unsyncronized views."""
        h_bar = self.viewRight.horizontalScrollBar()
        if h_bar and h_bar.isVisible():
            height = h_bar.height()
        else:
            height = 0
        if self.leftLayout.count() > 1:
            # remove previous spacer
            self.leftLayout.takeAt(1)
        if height:
            self.leftLayout.addSpacing(height)

    def closeEvent(self, unusedEvent:Optional[QEvent]) ->None:
        """update action button state"""
        assert Internal.mainWindow
        Internal.mainWindow.actionScoreTable.setChecked(False)


class ExplainView(QListView):

    """show a list explaining all score computations"""

    def __init__(self, scene:'ScoringScene') ->None:
        QListView.__init__(self)
        self.scene = scene
        decorateWindow(self, i18nc("@title:window", "Explain Scores").replace('&', ''))
        self.setGeometry(0, 0, 300, 400)
        self._model = QStringListModel()
        self.setModel(self._model)
        StateSaver(self)
        self.refresh()

    @property
    def game(self) ->'ScoringGame':
        """a proxy"""
        return self.scene.game

    def refresh(self) ->None:
        """refresh for new values"""
        lines = []
        if self.game is None:
            lines.append(i18n('There is no active game'))
        else:
            i18nName = i18n(self.game.ruleset.name)
            lines.append(i18n('%1', i18nName))
            lines.append('')
            for player in self.game.players:
                pLines = []
                explainHand = player.explainHand()
                if explainHand.hasTiles():
                    total = explainHand.total()
                    if total:
                        pLines = [f'{player.localName}: {total}']
                        for line in explainHand.explain():
                            pLines.append('- ' + line)
                elif player.handTotal:
                    pLines.append(
                        i18n(
                            'Manual score for %1: %2 points',
                            player.localName,
                            player.handTotal))
                if pLines:
                    pLines.append('')
                lines.extend(pLines)
        if 'xxx'.join(lines) != 'xxx'.join(str(x) for x in self._model.stringList()): # TODO: ohne?
            # QStringListModel does not optimize identical lists away, so we do
            self._model.setStringList(lines)

    def closeEvent(self, unusedEvent:Optional[QEvent]) ->None:
        """update action button state"""
        assert Internal.mainWindow
        Internal.mainWindow.actionExplain.setChecked(False)


class PenaltyBox(QSpinBox):

    """with its own validator, we only accept multiples of parties"""

    def __init__(self, parties:int, parent:Optional[QWidget]=None) ->None:
        super().__init__(parent)
        self.parties = parties
        self.prevValue = 0

    def validate(self, inputData:str, pos:int) -> Tuple[QValidator.State, str, int]:  # type:ignore[override]
        """check if value is a multiple of parties"""
        _ = super().validate(inputData, pos)
        result = _[0]
        inputData = _[1]
        newPos = _[2]
        if result == QValidator.State.Acceptable:
            try:
                int_data = int(inputData)
            except ValueError:
                return (QValidator.State.Invalid, inputData, pos)
            if int_data % self.parties != 0:
                result = QValidator.State.Intermediate
        if result == QValidator.State.Acceptable:
            self.prevValue = int(inputData)
        return (result, inputData, newPos)

    def fixup(self, data:str) ->str:  # type:ignore[override]
        """change input to a legal value"""
        # The Qt doc says return type is None but annotation says str
        value = int(str(data))
        prevValue = self.prevValue
        assert value != prevValue
        common = int(self.parties * 10)
        small = value // common
        if value > prevValue:
            newV = str(int((small + 1) * common))
        else:
            newV = str(int(small * common))
        return newV


class RuleBox(QCheckBox):

    """additional attribute: ruleId"""

    def __init__(self, rule:'Rule') ->None:
        super().__init__(i18n(rule.name))
        self.rule = rule

    def setApplicable(self, applicable:bool) ->None:
        """update box"""
        self.setVisible(applicable)
        if not applicable:
            self.setChecked(False)

class PlayerSelection(ListComboBox):

    """just add a label"""

    def __init__(self, items:List[Any], parent:Optional['QWidget']=None) ->None:
        super().__init__(items, parent)
        self.label = QLabel()


class PenaltyDialog(QDialog):

    """enter penalties"""

    def __init__(self, game:'ScoringGame', parent:Optional[QWidget]=None) ->None:
        """selection for this player, tiles are the still available tiles"""
        super().__init__(parent)
        decorateWindow(self, i18n("Penalty"))
        self.game = game
        grid = QGridLayout(self)
        lblOffense = QLabel(i18n('Offense:'))
        crimes = list(
            x for x in game.ruleset.penaltyRules if not ('absolute' in x.options and game.winner))
        self.cbCrime = ListComboBox(crimes)
        lblOffense.setBuddy(self.cbCrime)
        grid.addWidget(lblOffense, 0, 0)
        grid.addWidget(self.cbCrime, 0, 1, 1, 4)
        lblPenalty = QLabel(i18n('Total Penalty'))
        self.spPenalty = PenaltyBox(2)
        self.spPenalty.setRange(0, 9999)
        lblPenalty.setBuddy(self.spPenalty)
        self.lblUnits = QLabel(i18n('points'))
        grid.addWidget(lblPenalty, 1, 0)
        grid.addWidget(self.spPenalty, 1, 1)
        grid.addWidget(self.lblUnits, 1, 2)
        # a penalty can never involve the winner, neither as payer nor as payee, so max 3
        self.payers = list(PlayerSelection(game.losers()) for x in range(3))
        self.payees = list(PlayerSelection(game.losers()) for x in range(3))
        for idx, payer in enumerate(self.payers):
            grid.addWidget(payer, 3 + idx, 0)
            grid.addWidget(payer.label, 3 + idx, 1)
        for idx, payee in enumerate(self.payees):
            grid.addWidget(payee, 3 + idx, 3)
            grid.addWidget(payee.label, 3 + idx, 4)
        grid.addWidget(QLabel(''), 6, 0)
        grid.setRowStretch(6, 10)
        for player in self.payers + self.payees:
            player.currentIndexChanged.connect(self.playerChanged)
        self.spPenalty.valueChanged.connect(self.penaltyChanged)
        self.cbCrime.currentIndexChanged.connect(self.crimeChanged)
        buttonBox = KDialogButtonBox(self)
        grid.addWidget(buttonBox, 7, 0, 1, 5)
        buttonBox.setStandardButtons(QDialogButtonBox.StandardButton.Cancel)
        buttonBox.rejected.connect(self.reject)
        self.btnExecute = buttonBox.addButton(
            i18n("&Execute"),
            QDialogButtonBox.ButtonRole.AcceptRole)
        assert self.btnExecute
        self.btnExecute.clicked.connect(self.accept)
        self.crimeChanged()
        StateSaver(self)

    def accept(self) ->None:
        """execute the penalty"""
        offense = self.cbCrime.current
        payers = [x.current for x in self.payers if x.isVisible()]
        payees = [x.current for x in self.payees if x.isVisible()]
        for player in self.game.players:
            if player in payers:
                amount = -self.spPenalty.value() // len(payers)
            elif player in payees:
                amount = self.spPenalty.value() // len(payees)
            else:
                amount = 0
            player.getsPayment(amount)
            self.game.savePenalty(player, offense, amount)
        QDialog.accept(self)

    def usedCombos(self, partyCombos:List[PlayerSelection]) ->List[PlayerSelection]:
        """return all used player combos for this offense"""
        return [x for x in partyCombos if x.isVisibleTo(self)]

    def allParties(self) ->List['Player']:
        """return all parties involved in penalty payment"""
        return [cast('Player', x.current) for x in self.usedCombos(self.payers + self.payees)]

    def playerChanged(self) ->None:
        """shuffle players to ensure everybody only appears once.
        enable execution if all input is valid"""
        changedCombo = self.sender()
        if not isinstance(changedCombo, PlayerSelection):
            changedCombo = self.payers[0]
        usedPlayers = set(self.allParties())
        unusedPlayers = set(self.game.losers()) - usedPlayers
        foundPlayers = [cast('Player', changedCombo.current)]
        for combo in self.usedCombos(self.payers + self.payees):
            if combo is not changedCombo:
                if combo.current in foundPlayers:
                    combo.current = unusedPlayers.pop()
                foundPlayers.append(combo.current)

    def crimeChanged(self) ->None:
        """another offense has been selected"""
        offense = self.cbCrime.current
        payers = int(offense.options.get('payers', 1))
        payees = int(offense.options.get('payees', 1))
        self.spPenalty.prevValue = -offense.score.points
        self.spPenalty.setValue(-offense.score.points)
        self.spPenalty.parties = max(payers, payees)
        self.spPenalty.setSingleStep(10)
        self.lblUnits.setText(i18n('points'))
        self.playerChanged()
        self.penaltyChanged()

    def penaltyChanged(self) ->None:
        """total has changed, update payments"""
        # normally value is only validated when leaving the field
        self.spPenalty.interpretText()
        offense = self.cbCrime.current
        penalty = self.spPenalty.value()
        payers = int(offense.options.get('payers', 1))
        payees = int(offense.options.get('payees', 1))
        payerAmount = -penalty // payers
        payeeAmount = penalty // payees
        for pList, amount, count in ((self.payers, payerAmount, payers), (self.payees, payeeAmount, payees)):
            for idx, player in enumerate(pList):
                player.setVisible(idx < count)
                player.label.setVisible(idx < count)
                if idx < count:
                    if pList == self.payers:
                        player.label.setText(
                            i18nc(
                                'penalty dialog, appears behind paying player combobox',
                                'pays %1 points', -amount))
                    else:
                        player.label.setText(
                            i18nc(
                                'penalty dialog, appears behind profiting player combobox',
                                'gets %1 points', amount))


class ScoringDialog(QWidget):

    """a dialog for entering the scores"""
    # pylint: disable=too-many-instance-attributes

    def __init__(self, scene:'ScoringScene', parent:Optional[QWidget]=None) ->None:
        super().__init__(parent)
        self.scene = scene
        decorateWindow(self, i18nc("@title:window", "Scoring for this Hand"))
        self.nameLabels = list(QLabel() for x in range(4))
        self.spValues:List[QSpinBox] = list(QSpinBox() for x in range(4))
        self.windLabels = list(WindLabel() for x in range(4))
        self.wonBoxes = list(QCheckBox("") for x in range(4))
        self.detailsLayout = list(QVBoxLayout() for x in range(4))
        self.details = list(QWidget() for x in range(4))
        self.__meldPixMaps:List[QPixmap] = []
        grid = QGridLayout(self)
        pGrid = QGridLayout()
        grid.addLayout(pGrid, 0, 0, 2, 1)
        pGrid.addWidget(QLabel(i18nc('kajongg', "Player")), 0, 0)
        pGrid.addWidget(QLabel(i18nc('kajongg', "Wind")), 0, 1)
        pGrid.addWidget(QLabel(i18nc('kajongg', 'Score')), 0, 2)
        pGrid.addWidget(QLabel(i18n("Winner")), 0, 3)
        self.detailTabs = QTabWidget()
        self.detailTabs.setDocumentMode(True)
        pGrid.addWidget(self.detailTabs, 0, 4, 8, 1)
        for idx in range(4):
            self.setupUiForPlayer(pGrid, idx)
        self.draw = QCheckBox(i18nc('kajongg', 'Draw'))
        self.draw.clicked.connect(self.wonChanged)
        btnPenalties = QPushButton(i18n("&Penalties"))
        btnPenalties.clicked.connect(self.penalty)
        self.btnSave = QPushButton(i18n('&Save Hand'))
        self.btnSave.clicked.connect(self.game.nextScoringHand)
        self.btnSave.setEnabled(False)
        self.setupUILastTileMeld(pGrid)
        pGrid.setRowStretch(87, 10)
        pGrid.addWidget(self.draw, 7, 3)
        self.cbLastTile.currentIndexChanged.connect(self.slotLastTile)
        self.cbLastMeld.currentIndexChanged.connect(self.slotInputChanged)
        btnBox = QHBoxLayout()
        btnBox.addWidget(btnPenalties)
        btnBox.addWidget(self.btnSave)
        pGrid.addLayout(btnBox, 8, 4)
        StateSaver(self)
        self.refresh()

    @property
    def game(self) ->'ScoringGame':
        """proxy"""
        return self.scene.game

    def setupUILastTileMeld(self, pGrid:QGridLayout) ->None:
        """setup UI elements for last tile and last meld"""
        self.lblLastTile = QLabel(i18n('&Last Tile:'))
        self.cbLastTile = QComboBox()
        self.cbLastTile.setMinimumContentsLength(1)
        vpol = QSizePolicy()
        vpol.setHorizontalPolicy(QSizePolicy.Policy.Fixed)
        self.cbLastTile.setSizePolicy(vpol)
        self.cbLastTile.setSizeAdjustPolicy(
            QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)
        self.lblLastTile.setBuddy(self.cbLastTile)
        self.lblLastMeld = QLabel(i18n('L&ast Meld:'))
        self.prevLastTile:Optional['Tile'] = None
        self.cbLastMeld = QComboBox()
        self.cbLastMeld.setMinimumContentsLength(1)
        self.cbLastMeld.setSizePolicy(vpol)
        self.cbLastMeld.setSizeAdjustPolicy(
            QComboBox.SizeAdjustPolicy.AdjustToMinimumContentsLengthWithIcon)
        self.lblLastMeld.setBuddy(self.cbLastMeld)
        self.comboTilePairs:set['Tile'] = set()
        pGrid.setRowStretch(6, 5)
        pGrid.addWidget(self.lblLastTile, 7, 0, 1, 2)
        pGrid.addWidget(self.cbLastTile, 7, 2, 1, 1)
        pGrid.addWidget(self.lblLastMeld, 8, 0, 1, 2)
        pGrid.addWidget(self.cbLastMeld, 8, 2, 1, 2)

        self.lblLastTile.setVisible(False)
        self.cbLastTile.setVisible(False)
        self.cbLastMeld.setVisible(False)
        self.lblLastMeld.setVisible(False)

    def setupUiForPlayer(self, pGrid:QGridLayout, idx:int) ->None:
        """setup UI elements for a player"""
        self.spValues[idx] = QSpinBox()
        self.nameLabels[idx] = QLabel()
        self.nameLabels[idx].setBuddy(self.spValues[idx])
        self.windLabels[idx] = WindLabel()
        pGrid.addWidget(self.nameLabels[idx], idx + 2, 0)
        pGrid.addWidget(self.windLabels[idx], idx + 2, 1)
        pGrid.addWidget(self.spValues[idx], idx + 2, 2)
        pGrid.addWidget(self.wonBoxes[idx], idx + 2, 3)
        self.wonBoxes[idx].clicked.connect(self.wonChanged)
        self.spValues[idx].valueChanged.connect(self.slotInputChanged)
        detailTab = QWidget()
        self.detailTabs.addTab(detailTab, '')
        self.details[idx] = QWidget()
        detailTabLayout = QVBoxLayout(detailTab)
        detailTabLayout.addWidget(self.details[idx])
        detailTabLayout.addStretch()
        self.detailsLayout[idx] = QVBoxLayout(self.details[idx])

    def refresh(self) ->None:
        """reload game"""
        self.clear()
        game = self.game
        self.setVisible(game is not None and not game.finished())
        if game:
            for idx, player in enumerate(game.players):
                for child in self.details[idx].children():
                    if isinstance(child, RuleBox):
                        child.hide()
                        self.detailsLayout[idx].removeWidget(child)
                        del child
                if game:
                    self.spValues[idx].setRange(0, int(game.ruleset.limit) or 99999)
                    self.nameLabels[idx].setText(player.localName)
                    self.refreshWindLabels()
                    self.detailTabs.setTabText(idx, player.localName)
                    player.manualRuleBoxes = [RuleBox(x)
                                              for x in game.ruleset.allRules if x.hasSelectable]
                    for ruleBox in player.manualRuleBoxes:
                        self.detailsLayout[idx].addWidget(ruleBox)
                        ruleBox.clicked.connect(self.slotInputChanged)
                player.refreshManualRules()

    def show(self) ->None:
        """only now compute content"""
        if self.game and not self.game.finished():
            self.slotInputChanged()
            QWidget.show(self)

    def penalty(self) ->None:
        """penalty button clicked"""
        dlg = PenaltyDialog(self.game)
        dlg.exec()

    def slotLastTile(self) ->None:
        """called when the last tile changes"""
        newLastTile = self.computeLastTile()
        if not newLastTile:
            return
        if self.prevLastTile and self.prevLastTile.isExposed != newLastTile.isExposed:
            # state of last tile (concealed/exposed) changed:
            # for all checked boxes check if they still are applicable
            winner = cast('ScoringPlayer', self.game.winner)
            if winner:
                for box in winner.manualRuleBoxes:
                    if box.isChecked():
                        box.setChecked(False)
                        if winner.hand.manualRuleMayApply(box.rule):
                            box.setChecked(True)
        self.prevLastTile = newLastTile
        self.fillLastMeldCombo()
        self.slotInputChanged()

    def computeLastTile(self) ->Tile:
        """return the currently selected last tile"""
        idx = self.cbLastTile.currentIndex()
        if idx >= 0:
            return self.cbLastTile.itemData(idx)
        return Tile.none

    def clickedPlayerIdx(self, checkbox:'QObject') ->int:
        """the player whose box has been clicked"""
        for idx in range(4):
            if checkbox == self.wonBoxes[idx]:
                return idx
        assert False
        return 0

    def wonChanged(self) ->None:
        """if a new winner has been defined, uncheck any previous winner"""
        newWinner = None
        if sender := self.sender():
            if sender != self.draw:
                clicked = self.clickedPlayerIdx(sender)
                if self.wonBoxes[clicked].isChecked():
                    newWinner = self.game.players[clicked]
                else:
                    newWinner = None
        self.game.winner = newWinner
        for idx in range(4):
            if newWinner != self.game.players[idx]:
                self.wonBoxes[idx].setChecked(False)
        if newWinner:
            self.draw.setChecked(False)
            self.lblLastTile.setVisible(True)
            self.cbLastTile.setVisible(True)
        self.lblLastMeld.setVisible(False)
        self.cbLastMeld.setVisible(False)
        self.fillLastTileCombo()
        self.slotInputChanged()

    def updateManualRules(self) ->None:
        """enable/disable them"""
        # if an exclusive rule has been activated, deactivate it for
        # all other players
        ruleBox = self.sender()
        if isinstance(ruleBox, RuleBox) and ruleBox.isChecked() and ruleBox.rule.exclusive():
            for idx, player in enumerate(self.game.players):
                if ruleBox.parentWidget() != self.details[idx]:
                    for pBox in player.manualRuleBoxes:
                        if pBox.rule.name == ruleBox.rule.name:
                            pBox.setChecked(False)
        newState = False
        if self.game.winner and self.game.winner.handBoard:
            newState = self.game.winner.handBoard.empty
        self.lblLastTile.setVisible(newState)
        self.cbLastTile.setVisible(newState)
        if self.game:
            for player in self.game.players:
                player.refreshManualRules(self.sender())

    def clear(self) ->None:
        """prepare for next hand"""
        if self.game:
            for idx, player in enumerate(self.game.players):
                self.spValues[idx].clear()
                self.spValues[idx].setValue(0)
                self.wonBoxes[idx].setChecked(False)
                player.payment = 0
                player.invalidateHand()
        for box in self.wonBoxes:
            box.setVisible(False)
        self.draw.setChecked(False)
        self.updateManualRules()

        if self.game is None:
            self.hide()
        else:
            self.refreshWindLabels()
            self.computeScores()
            self.spValues[0].setFocus()
            self.spValues[0].selectAll()

    def refreshWindLabels(self) ->None:
        """update their wind and prevailing"""
        for idx, player in enumerate(self.game.players):
            self.windLabels[idx].wind = player.wind
            self.windLabels[idx].prevailing = self.game.point.prevailing

    def computeScores(self) ->None:
        """if tiles have been selected, compute their value"""
        # too many branches
        if not self.game:
            return
        if self.game.finished():
            self.hide()
            return
        for nameLabel, wonBox, spValue, player in zip(
                self.nameLabels, self.wonBoxes, self.spValues, self.game.players):
            with BlockSignals([spValue, wonBox]):
                # we do not want that change to call computeScores again
                if player.handBoard and not player.handBoard.empty:
                    spValue.setEnabled(False)
                    nameLabel.setBuddy(wonBox)
                    for _ in range(10):
                        prevTotal = player.handTotal
                        player.invalidateHand()
                        wonBox.setVisible(player.hand.won)
                        if not wonBox.isVisibleTo(self) and wonBox.isChecked():
                            wonBox.setChecked(False)
                            self.game.winner = None
                        elif prevTotal == player.handTotal:
                            break
                        player.refreshManualRules()
                    spValue.setValue(player.handTotal)
                else:
                    if not spValue.isEnabled():
                        spValue.clear()
                        spValue.setValue(0)
                        spValue.setEnabled(True)
                        nameLabel.setBuddy(spValue)
                    wonBox.setVisible(
                        player.handTotal >= self.game.ruleset.minMJTotal())
                    if not wonBox.isVisibleTo(self) and wonBox.isChecked():
                        wonBox.setChecked(False)
                if not wonBox.isVisibleTo(self) and player is self.game.winner:
                    self.game.winner = None
        assert Internal.scene
        if Internal.scene.explainView:
            Internal.scene.explainView.refresh()

    def __lastMeldContent(self) ->Tuple[set[Tile], List['UITile']]:
        """prepare content for lastmeld combo"""
        lastTiles = set()
        winnerTiles = []
        if self.game.winner and self.game.winner.handBoard:
            winnerTiles = self.game.winner.handBoard.uiTiles
            pairs = []
            for meld in self.game.winner.hand.melds:
                if len(meld) < 4:
                    pairs.extend(meld)
            for tile in winnerTiles:
                if tile.tile in pairs and not tile.isBonus:
                    lastTiles.add(tile.tile)
        return lastTiles, winnerTiles

    def __fillLastTileComboWith(self, lastTiles:set[Tile], winnerTiles:List['UITile']) ->None:
        """fill last meld combo with prepared content"""
        self.comboTilePairs = lastTiles
        idx = max(self.cbLastTile.currentIndex(), 0)
        indexedTile = self.cbLastTile.itemData(idx)
        restoredIdx = None
        self.cbLastTile.clear()
        if not winnerTiles:
            return
        assert winnerTiles[0].board
        pmSize = winnerTiles[0].board.tileset.faceSize.toSize()
        self.cbLastTile.setIconSize(pmSize)
        QPixmapCache.clear()
        shownTiles = set()
        for tile in winnerTiles:
            if tile.tile in lastTiles and tile.tile not in shownTiles:
                shownTiles.add(tile.tile)
                self.cbLastTile.addItem(
                    QIcon(tile.pixmapFromSvg(pmSize, withBorders=False)),
                    '', tile.tile)
                if indexedTile is tile.tile:
                    restoredIdx = self.cbLastTile.count() - 1
        if not restoredIdx and indexedTile:
            # try again, maybe the tile changed between concealed and exposed
            indexedTile = indexedTile.exposed
            for idx in range(self.cbLastTile.count()):
                if indexedTile is self.cbLastTile.itemData(idx).exposed:
                    restoredIdx = idx
                    break
        if not restoredIdx:
            restoredIdx = 0
        self.cbLastTile.setCurrentIndex(restoredIdx)
        self.prevLastTile = self.computeLastTile()

    def clearLastTileCombo(self) ->None:
        """as the name says"""
        self.comboTilePairs = set()
        self.cbLastTile.clear()

    def fillLastTileCombo(self) ->None:
        """fill the drop down list with all possible tiles.
        If the drop down had content before try to preserve the
        current index. Even if the tile changed state meanwhile."""
        if self.game is None:
            return
        lastTiles, winnerTiles = self.__lastMeldContent()
        if self.comboTilePairs == lastTiles:
            return
        with BlockSignals([self.cbLastTile]):
            # we only want to emit the changed signal once
            self.__fillLastTileComboWith(lastTiles, winnerTiles)
        self.cbLastTile.currentIndexChanged.emit(0)

    def __fillLastMeldComboWith(self, winnerMelds:MeldList, indexedMeld:'Meld', lastTile:Optional[Tile]) ->None:
        """fill last meld combo with prepared content"""
        winner = self.game.winner
        assert winner
        assert winner.handBoard
        faceWidth = int(winner.handBoard.tileset.faceSize.width())
        faceHeight = int(winner.handBoard.tileset.faceSize.height())
        restoredIdx = None
        for meld in winnerMelds:
            pixMap = QPixmap(faceWidth * len(meld), faceHeight)
            pixMap.fill(Qt.GlobalColor.transparent)
            self.__meldPixMaps.append(pixMap)
            painter = QPainter(pixMap)
            for element in meld:
                painter.drawPixmap(0, 0,
                                   winner.handBoard.tilesByElement(element)
                                   [0].pixmapFromSvg(QSize(faceWidth, faceHeight), withBorders=False))
                painter.translate(QPointF(faceWidth, 0.0))
            self.cbLastMeld.addItem(QIcon(pixMap), '', meld)
            if indexedMeld == meld:
                restoredIdx = self.cbLastMeld.count() - 1
        if not restoredIdx and indexedMeld:
            # try again, maybe the meld changed between concealed and exposed
            indexedMeld = indexedMeld.exposed
            for idx in range(self.cbLastMeld.count()):
                meldContent = self.cbLastMeld.itemData(idx)
                if indexedMeld == meldContent.exposed:
                    restoredIdx = idx
                    if lastTile is not None and lastTile not in meldContent:
                        lastTile = lastTile.swapped
                        assert lastTile in meldContent
                        with BlockSignals([self.cbLastTile]):  # we want to continue right here
                            idx = self.cbLastTile.findData(lastTile)
                            self.cbLastTile.setCurrentIndex(idx)
                    break
        if not restoredIdx:
            restoredIdx = 0
        self.cbLastMeld.setCurrentIndex(restoredIdx)
        self.cbLastMeld.setIconSize(QSize(faceWidth * 3, faceHeight))

    def fillLastMeldCombo(self) ->None:
        """fill the drop down list with all possible melds.
        If the drop down had content before try to preserve the
        current index. Even if the meld changed state meanwhile."""
        with BlockSignals([self.cbLastMeld]):  # we only want to emit the changed signal once
            showCombo = False
            idx = max(self.cbLastMeld.currentIndex(), 0)
            indexedMeld = self.cbLastMeld.itemData(idx)
            self.cbLastMeld.clear()
            self.__meldPixMaps = []
            if not self.game.winner:
                return
            if self.cbLastTile.count() == 0:
                return
            assert Internal.scene
            lastTile = cast('ScoringScene', Internal.scene).computeLastTile()
            winnerMelds = MeldList(m for m in self.game.winner.hand.melds if len(m) < 4
                           and lastTile in m)
            assert winnerMelds, f'lastTile {lastTile} missing in {self.game.winner.hand.melds}'
            if len(winnerMelds) == 1:
                self.cbLastMeld.addItem(QIcon(), '', winnerMelds[0])
                self.cbLastMeld.setCurrentIndex(0)
                self.lblLastMeld.setVisible(False)
                self.cbLastMeld.setVisible(False)
                return
            showCombo = True
            self.__fillLastMeldComboWith(winnerMelds, indexedMeld, lastTile)
            self.lblLastMeld.setVisible(showCombo)
            self.cbLastMeld.setVisible(showCombo)
        self.cbLastMeld.currentIndexChanged.emit(0)

    def slotInputChanged(self) ->None:
        """some input fields changed: update"""
        for player in self.game.players:
            player.invalidateHand()
        self.updateManualRules()
        self.computeScores()
        self.validate()
        for player in self.game.players:
            player.showInfo()
        assert Internal.mainWindow
        Internal.mainWindow.updateGUI()

    def validate(self) ->None:
        """update the status of the OK button"""
        game = self.game
        if game:
            valid = True
            if game.winner and game.winner.handTotal < game.ruleset.minMJTotal():
                valid = False
            elif not game.winner and not self.draw.isChecked():
                valid = False
            self.btnSave.setEnabled(valid)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2017 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0

Start this in the installation directory of Kajongg: That
is where this program resides. Below you find a code
block that might have to be adapted.
"""

from distutils.core import setup
from distutils.command.build import build
from subprocess import call

import re
import os

# Adapt this range: =======================================================
FULLAUTHOR = "Wolfgang Rohdewald <wolfgang@rohdewald.de>"
LICENSE = 'GNU General Public License v2'
URL = "https://apps.kde.org/kajongg"
try:
    from appversion import VERSION
except ImportError:
    VERSION = "Unknown"

# =======================================================

# This most certainly does not run on Windows. We do not care for now.
# at least all / in paths would have to be changed

(AUTHOR, EMAIL) = re.match('^(.*?)\s*<(.*)>$', FULLAUTHOR).groups()

os.umask(0o0022) # files should be readable and executable by everybody

kdeDirs = {}
for type in 'exe', 'data', 'xdgdata-apps', 'icon', 'html':
    kdeDirs[type] = os.popen("kf5-config --expandvars --install %s" % type).read().strip()
kdeDirs['iconApps'] = os.path.join(kdeDirs['icon'], 'hicolor', 'scalable', 'apps')
kdeDirs['iconActions'] = os.path.join(kdeDirs['icon'], 'hicolor', 'scalable', 'actions')

app_files = [os.path.join('src', x) for x in os.listdir('src') if x.endswith('.py')]
app_files.append('src/kajonggui.rc')
app_files.append('COPYING')
app_files.append('COPYING.DOC')

doc_files = [os.path.join('doc', x) for x in os.listdir('doc') if x.endswith('.png')]

for ignFile in os.listdir('src'):
    if ignFile.endswith('.pyc'):
        os.remove(os.path.join('src', ignFile))

data_files = [ \
    (kdeDirs['exe'], ['kajongg', 'kajonggserver']),
    (os.path.join(kdeDirs['data'], 'kajongg'), app_files),
    (os.path.join(kdeDirs['html'], 'en', 'kajongg'), doc_files),
    (kdeDirs['xdgdata-apps'], ['org.kde.kajongg.desktop']),
    ('/usr/share/doc/kajongg/', ['COPYING.DOC']),
    (kdeDirs['iconApps'], ['kajongg.svgz']),
    (kdeDirs['iconActions'], ['games-kajongg-law.svgz'])]

voice_directories = [x for x in os.listdir('voices') if x.startswith('male') or x.startswith('female')]
for directory in voice_directories:
    data_files.append((os.path.join(kdeDirs['data'], 'kajongg', 'voices', directory), [os.path.join('voices', directory, x) for x in os.listdir(os.path.join('voices', directory))]))

extra = {}
# extra['requires'] = ('pyQt4', 'sdf') does not do anything

class KmjBuild(build):

    def run(self):
        for binary in ['kajongg', 'kajonggserver']:
            open(binary, 'w').write('#!/bin/sh\nexec %skajongg/%s.py $*\n' % (kdeDirs['data'], binary))
            os.chmod(binary, 0o0755)
        call(['cp sc-apps-kajongg.svgz kajongg.svgz'], shell=True)
        call(['cp sc-action-games-kajongg-law.svgz games-kajongg-law.svgz'], shell=True)
        build.run(self)


setup(name='kajongg',
    version=VERSION,
    description='The classical game of Mah Jongg',
    long_description="This is the classical Mah Jongg for four players. "
            "If you are looking for the Mah Jongg solitaire please use the "
            "application kmahjongg.",
    author=AUTHOR,
    author_email=EMAIL,
    url=URL,
    download_url='https://www.linux-apps.com/p/1109453/',
    data_files=data_files,
    cmdclass = { 'build' : KmjBuild },  # define custom build class
    license=LICENSE,
    classifiers=['Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: OS Independent',
        'Environment :: X11 Applications :: KDE',
        'Topic :: Desktop Environment :: K Desktop Environment (KDE)',
        'Topic :: Games:: Board Games',
        'Intended Audience :: End Users/Desktop',
        'Programming Language :: Python',],
    **extra)



# -*- coding: utf-8 -*-

"""
Copyright (C) 2008-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

from typing import Optional, TYPE_CHECKING, Dict, Any
import sqlite3

from mi18n import i18n, i18nc
from kde import KIcon
from dialogs import Sorry, QuestionYesNo
from qt import Qt
from qt import QDialog, QHBoxLayout, QVBoxLayout, QDialogButtonBox, \
    QTableWidget, QTableWidgetItem, QWidget

from common import Internal
from query import Query
from guiutil import decorateWindow
from statesaver import StateSaver

if TYPE_CHECKING:
    from qt import QEvent, QKeyEvent


class PlayerList(QDialog):

    """QtSQL Model view of the players"""

    def __init__(self, parent:Optional['QWidget']=None):
        super().__init__(parent)
        self._data:Dict[str, int] = {}
        self.table = QTableWidget(self)
        if header := self.table.horizontalHeader():
            header.setStretchLastSection(True)
        if header := self.table.verticalHeader():
            header.setVisible(False)
        self.table.setEditTriggers(QTableWidget.EditTrigger.NoEditTriggers)
        self.table.itemChanged.connect(self.itemChanged)
        self.updateTable()
        self.buttonBox = QDialogButtonBox()
        self.buttonBox.setStandardButtons(
            QDialogButtonBox.StandardButton.Close)  # Close has the Rejected role
        self.buttonBox.rejected.connect(self.accept)
        self.newButton = self.buttonBox.addButton(
            i18nc('define a new player',
                  "&New"),
            QDialogButtonBox.ButtonRole.ActionRole)
        if button := self.newButton:
            button.setIcon(KIcon("document-new"))
            button.clicked.connect(self.slotInsert)
        self.deleteButton = self.buttonBox.addButton(
            i18n("&Delete"), QDialogButtonBox.ButtonRole.ActionRole)
        assert self.deleteButton
        self.deleteButton.setIcon(KIcon("edit-delete"))
        self.deleteButton.clicked.connect(self.delete)

        cmdLayout = QHBoxLayout()
        cmdLayout.addWidget(self.buttonBox)
        layout = QVBoxLayout()
        layout.addWidget(self.table)
        layout.addLayout(cmdLayout)
        self.setLayout(layout)
        decorateWindow(self, i18n("Players"))
        self.setObjectName('Players')

    def showEvent(self, unusedEvent:Optional['QEvent']) ->None:
        """adapt view to content"""
        StateSaver(self, self.table)

    @staticmethod
    def sortKey(text:str) ->str:
        """display order in Table"""
        return text.upper() if text else 'zzzzzzzzzzzz'

    def updateTable(self, data:Optional[Dict[str, int]]=None, currentName:Optional[str]=None) ->None:
        """fills self.table from DB"""
        self.table.itemChanged.disconnect(self.itemChanged)
        table = self.table
        table.clear()
        if data is None:
            data = dict(
                Query('select name, id from player where name not like "ROBOT %"').records)
        self._data = data
        table.setColumnCount(1)
        table.setRowCount(len(self._data))
        table.setHorizontalHeaderLabels([i18n("Player")])
        table.setSelectionBehavior(QTableWidget.SelectionBehavior.SelectRows)
        table.setSelectionMode(QTableWidget.SelectionMode.SingleSelection)
        selectedItem = None
        for row, name in enumerate(sorted(self._data, key=self.sortKey)):
            item = QTableWidgetItem(name)
            if selectedItem is None:
                selectedItem = item
            table.setItem(row, 0, item)
            if name == currentName:
                selectedItem = item
        if selectedItem:
            table.setCurrentItem(selectedItem)
            table.scrollToItem(selectedItem)
        self.table.itemChanged.connect(self.itemChanged)

    def itemChanged(self, item:QTableWidgetItem) ->None:
        """this must be new because editing is disabled for others"""
        currentName = item.text()
        if currentName in self._data:
            Sorry(i18n('Player %1 already exists', currentName))
            self.setFocus()
            if _ := self.table.item(self.table.currentRow(), 0):
                del self._data[_.text()]
            self.updateTable(currentName=currentName)
            return
        try:
            Query('insert into player(name) values(?)', (currentName, ))
        except sqlite3.Error as exc:
            Sorry(
                i18n(
                    'Error while adding player %1: %2',
                    currentName,
                    str(exc)))
        self.updateTable(currentName=currentName)

    def slotInsert(self) ->None:
        """insert a record"""
        self._data[''] = 0
        self.updateTable(data=self._data, currentName='')
        for row in range(len(self._data)):
            if item := self.table.item(row, 0):
                if not item.text():
                    self.table.editItem(item)

    @staticmethod
    def __deletePlayer(playerId:int) ->None:
        """delete this player and all associated games"""
        with Internal.db: # transaction
            Query("delete from score where player=?", (playerId, ))
            Query("delete from game where p0=? or p1=? or p2=? or p3=?", (playerId, ) * 4)
            Query("delete from player where id=?", (playerId,))

    def delete(self) ->None:
        """delete selected entry"""
        def answered(result:Any) ->None:
            """coming from QuestionYesNo"""
            if result is True:
                self.__deletePlayer(playerId)
            cleanup()
        def cleanup() ->None:
            """update table view"""
            self.updateTable()
            self.table.setCurrentCell(min(currentRow, len(self._data) - 1), 0)
            # the main window gets focus after QuestionYesNo
            self.activateWindow()

        items = self.table.selectedItems()
        if not items:
            return
        currentRow = self.table.currentRow()
        assert len(items) == 1
        name = items[0].text()
        playerId = self._data[name]

        fullCount = int(Query(
            "select count(1) from game where p0=? or p1=? or p2=? or p3=?",
            (playerId, ) * 4).records[0][0])
        if not fullCount:
            self.__deletePlayer(playerId)
            cleanup()
        else:
            finishedCount = int(Query(
                "select count(1) from game where (p0=? or p1=? or p2=? or p3=?) and endtime is not null",
                (playerId, ) * 4).records[0][0])
            QuestionYesNo(i18n(
                'There are %1 finished and %2 unfinished games for %3, delete %3 anyway?'
                '  This will also delete all games played by %3!',
                finishedCount, fullCount - finishedCount, name)).addBoth(answered)

    def keyPressEvent(self, event:Optional['QKeyEvent']) ->None:
        """use insert/delete keys for insert/delete"""
        if event:
            key = event.key()
            if key == Qt.Key.Key_Insert:
                self.slotInsert()
            elif key == Qt.Key.Key_Delete:
                self.delete()
            else:
                super().keyPressEvent(event)



# -*- coding: utf-8 -*-

"""
Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only

"""

import datetime
import weakref
from types import ModuleType
from typing import Tuple, Optional, List, Type, Any, TYPE_CHECKING, Union, cast

from twisted.spread import pb
from twisted.internet.task import deferLater
from twisted.internet.defer import Deferred, succeed, fail
from twisted.python.failure import Failure
from util import Duration
from log import logDebug, logException, logWarning, logFailure
from mi18n import i18nc
from message import Message
from common import Internal, Debug, Options, ReprMixin
from common import isAlive
from tilesource import TileSource
from rule import Ruleset
from game import PlayingGame
from query import Query
from move import Move
from animation import animate, animateAndDo
from player import PlayingPlayer

import intelligence
import altint


if TYPE_CHECKING:
    from deferredutil import Request
    from login import Connection
    from game import Game
    from tables import TableList
    from chat import ChatWindow
    from servertable import ServerTable
    from wind import Wind
    from scene import PlayingScene
    from message import ClientMessage, ServerMessage

class Table(ReprMixin):

    """defines things common to both ClientTable and ServerTable"""

    def __init__(self, tableid:int, ruleset:Union[Ruleset, str], suspendedAt:Optional[str],
                 running:bool, playOpen:bool, autoPlay:bool, wantedGame:str) ->None:
        self.tableid = tableid
        if isinstance(ruleset, Ruleset):
            self.ruleset = ruleset
        else:
            self.ruleset = Ruleset.cached(ruleset)
        self.suspendedAt = suspendedAt
        self.running = running
        self.playOpen = playOpen
        self.autoPlay = autoPlay
        self.wantedGame = wantedGame

    def status(self) ->str:
        """a status string"""
        result = ''
        if self.suspendedAt:
            result = i18nc('table status', 'Suspended')
            result += ' ' + datetime.datetime.strptime(
                self.suspendedAt,
                '%Y-%m-%dT%H:%M:%S').strftime('%c')
        if self.running:
            result += ' ' + i18nc('table status', 'Running')
        return result or i18nc('table status', 'New')

    def __str__(self) ->str:
        return f'Table({self.tableid})'


class ClientTable(Table):

    """the table as seen by the client"""
    # pylint: disable=too-many-arguments

    def __init__(self, client:'Client', tableid:int, ruleset:Ruleset, gameid:int,  # pylint:disable=too-many-positional-arguments
                 suspendedAt:str, running:bool,
                 playOpen:bool, autoPlay:bool, wantedGame:str, playerNames:List[str],
                 playersOnline:List[PlayingPlayer], endValues:Tuple[List[int], List[int]]) ->None:
        Table.__init__(
            self,
            tableid,
            ruleset,
            suspendedAt,
            running,
            playOpen,
            autoPlay,
            wantedGame)
        self.client = client
        self.gameid = gameid
        self.playerNames = playerNames
        self.playersOnline = playersOnline
        self.endValues = endValues
        self.myRuleset = None  # if set, points to an identical local ruleset
        for myRuleset in Ruleset.availableRulesets():
            if myRuleset == self.ruleset:
                self.myRuleset = myRuleset
                break
        self.chatWindow:Optional['ChatWindow'] = None

    def isOnline(self, player:Optional[str]) ->bool:  # TODO: why optional?
        """did he join the tabled?"""
        for idx, name in enumerate(self.playerNames):
            if player == name:
                return bool(self.playersOnline[idx])
        return False

    def __str__(self) ->str:
        onlineNames = [x for x in self.playerNames if self.isOnline(x)]
        offlineString = ''
        offlineNames = [
            x for x in self.playerNames if x not in onlineNames
            and not x.startswith('Robot')]
        if offlineNames:
            offlineString = ' offline:' + ','.join(offlineNames)
        return f"{int(self.tableid)}({self.ruleset.name} {','.join(onlineNames)}{offlineString})"

    def gameExistsLocally(self) ->bool:
        """True if the game exists in the data base of the client"""
        assert self.gameid
        return bool(Query('select 1 from game where id=?', (self.gameid,)).records)


class Client(pb.Referenceable):

    """interface to the server. This class only implements the logic,
    so we can also use it on the server for robot clients. Compare
    with HumanClient(Client)"""

    def __init__(self, name:Optional[str]=None) ->None:
        """name is something like Robot 1 or None for the game server"""
        self.name = name
        self.game:Optional['PlayingGame'] = None
        self.__connection:Optional['Connection'] = None
        self.tables:List[ClientTable] = []
        self._table:Union[ClientTable, 'ServerTable', None] = None
        self.tableList:Optional['TableList'] = None
        self.voiceId:Optional[str]  # only for mypy in servertable.py

    @property
    def table(self) ->Union[ClientTable, 'ServerTable', None]:
        """hide weakref"""
        if self._table:
            return self._table()  # type:ignore
        return None

    @table.setter
    def table(self, value:Union[ClientTable, 'ServerTable', None]) ->None:
        """hide weakref"""
        if value is not None:
            self._table = weakref.ref(value)  # type:ignore

    @property
    def connection(self) ->Optional['Connection']:
        """update main window title if needed"""
        return self.__connection

    @connection.setter
    def connection(self, value:'Connection') ->None:
        """update main window title if needed"""
        if self.__connection != value:
            self.__connection = value
            if Internal.scene:
                Internal.scene.mainWindow.updateGUI()

    def _tableById(self, tableid:int) ->Optional[ClientTable]:
        """return table with tableid"""
        for table in self.tables:
            if table.tableid == tableid:
                return table
        return None

    def logout(self, unusedResult:Optional[List['Request']]=None) ->Deferred:
        """virtual"""
        return succeed(None)

    def isRobotClient(self) ->bool:
        """avoid using isinstance because that imports too much for the server"""
        return bool(self.name)

    @staticmethod
    def isHumanClient() ->bool:
        """avoid using isinstance because that imports too much for the server"""
        return False

    def isServerClient(self) ->bool:
        """avoid using isinstance because that imports too much for the server"""
        return bool(not self.name)

    def remote_newTables(self, tables:List[Any]) ->None:
        """update table list"""
        newTables = [ClientTable(self, *x) for x in tables]
        self.tables.extend(newTables)
        if Debug.table:
            _ = ', '.join(str(ClientTable(self, *x)) for x in tables)
            logDebug(f'{self.name} got new tables:{_}')

    @staticmethod
    def remote_serverRulesets(hashes:List[str]) ->List[str]:
        """the server will normally send us hashes of rulesets. If
        a hash is not known by us, tell the server so it will send the
        full ruleset definition instead of the hash. It would be even better if
        the server always only sends the hash and the client then says "I do
        not know this ruleset, please send definition", but that would mean
        more changes to the client code"""
        return [x for x in hashes if not Ruleset.hashIsKnown(x)]

    def tableChanged(self, table:Any) ->Tuple[Optional[ClientTable], ClientTable]:
        """update table list"""
        newTable = ClientTable(self, *table)  # type: ignore
        oldTable = self._tableById(newTable.tableid)
        if oldTable:
            self.tables.remove(oldTable)
            self.tables.append(newTable) # this indent wants a comment
        return oldTable, newTable

    def remote_tableRemoved(self, tableid:int, message:str, *args:Any) ->None:  # pylint: disable=unused-argument
        """update table list"""
        table = self._tableById(tableid)
        if table:
            self.tables.remove(table)

    def reserveGameId(self, gameid:int) ->Message:
        """the game server proposes a new game id. We check if it is available
        in our local data base - we want to use the same gameid everywhere"""
        with Internal.db:
            assert self.connection
            query = Query('insert into game(id,seed) values(?,?)',
                          (gameid, self.connection.url), mayFail=True, failSilent=True)
            if query.rowcount() != 1:
                return Message.NO
        return Message.OK

    @staticmethod
    def __findAI(modules:List[ModuleType], aiName:str) ->Optional[Type]:
        """list of all alternative AIs defined in altint.py"""
        for module in modules:
            for key, value in module.__dict__.items():
                if key == 'AI' + aiName:
                    return value
        return None

    def __assignIntelligence(self) ->None:
        """assign intelligence to myself. All players already have default intelligence."""
        if self.isHumanClient():
            assert self.game
            aiClass = self.__findAI([intelligence, altint], Options.AI)
            if not aiClass:
                raise ValueError(f'intelligence {Options.AI} is undefined')
            self.game.myself.intelligence = aiClass(self.game.myself)

    def readyForGameStart(
            self, tableid:int, gameid:int, wantedGame:str, playerNames:List[Tuple['Wind', str]],
            shouldSave:bool=True, gameClass:Optional[Type]=None) ->Deferred:
        """the game server asks us if we are ready. A robot is always ready."""
        def disagree(about:str) ->None:
            """do not bother to translate this, it should normally not happen"""
            assert self.game  # mypy should be able to infer this
            self.game.close()
            msg = f'The data bases for game {self.game.seed} have different {about}'
            logWarning(msg)
            raise pb.Error(msg)
        if not self.table:
            assert not self.isRobotClient()
            self.table = self._tableById(tableid)
        else:
            assert self.isRobotClient()
            # robot client instance: self.table is already set
        assert self.table
        if gameClass is None:
            gameClass = PlayingGame
        if self.table.suspendedAt:
            game = cast(PlayingGame, gameClass.loadFromDB(gameid, self))
            self.game = game
            game.assignPlayers(playerNames)
            table = cast(ClientTable, self.table)
            if self.isHumanClient():
                if game.point.handCount != table.endValues[0]:
                    disagree(f'numbers for played hands: Server:{table.endValues[0]}, Client:{game.point.handCount}')
                for player in game.players:
                    if player.balance != table.endValues[1][player.wind.char]:
                        disagree(f'balances for wind {player.wind}: '
                                 f'Server:{table.endValues[1][player.wind]}, Client:{player.balance}')
        else:
            self.game = gameClass(playerNames, self.table.ruleset,
                                  gameid=gameid, wantedGame=wantedGame, client=self,
                                  playOpen=self.table.playOpen, autoPlay=self.table.autoPlay)
            assert self.game, f'cannot initialize game {gameid}'
        self.game.shouldSave = shouldSave
        self.__assignIntelligence()
                                  # intelligence variant is not saved for
                                  # suspended games
        self.game.prepareHand()
        return succeed(Message.OK)

    def readyForHandStart(self, playerNames:List[Tuple['Wind', str]],  # pylint:disable=useless-return,useless-suppression
        mustRotateWinds:bool) ->Optional[Deferred]:
        """the game server asks us if we are ready. A robot is always ready..."""
        if self.game:
            self.game.assignPlayers(playerNames)
            if mustRotateWinds:
                self.game.rotateWinds()
                if self.game.finished():
                    self.game.finish_in_db()
            self.game.prepareHand()
        # mypy 1.5.1 wants an explicit return
        return None

    def __delayAnswer(self, result:List['Request'], delay:float, delayStep:float) ->Any:
        """try again, may we chow now?"""
        if not self.game:
            # game has been aborted meanwhile
            return result
        noClaimCount = 0
        delay += delayStep
        for move in self.game.lastMoves():
            # latest move first
            if move.message == Message.Discard:
                break
            if move.message == Message.NoClaim and move.notifying:
                noClaimCount += 1
                if noClaimCount == 2:
                    if Debug.delayChow and self.game.lastDiscard:
                        self.game.debug(f'everybody said "I am not interested", so {self.game.myself.name} '
                                        f'claims chow now for {self.game.lastDiscard.name()}')
                    return result
            elif move.message in (Message.Pung, Message.Kong, Message.MahJongg) and move.notifying:
                if Debug.delayChow and self.game.lastDiscard:
                    self.game.debug(f'{move.player} said {move.message} so {self.game.myself} '
                                    f'suppresses Chow for {self.game.lastDiscard.name()}'.replace('  ', ' '))
                return Message.NoClaim
        if delay < self.game.ruleset.claimTimeout * 0.95:
            # one of those slow humans is still thinking
            return deferLater(Internal.reactor, delayStep, self.__delayAnswer, result, delay, delayStep)
        if Debug.delayChow and self.game.lastDiscard:
            self.game.debug(f'{self.game.myself.name} must chow now '
                            f'for {self.game.lastDiscard.name()} because timeout is over')
        return result

    def ask(self, move:Move, answers:List['ClientMessage']) ->Deferred:
        """place the robot AI here.
        send answer and one parameter to server"""
        delay = 0.0
        delayStep = 0.1
        assert self.game
        myself = self.game.myself
        cast(PlayingPlayer, myself).computeSayable(move, answers)
        result = myself.intelligence.selectAnswer(answers)
        assert result
        if result[0] == Message.Chow:
            if Debug.delayChow and self.game.lastDiscard:
                self.game.debug(f'{self.game.myself.name} waits to see if somebody '
                                f'says Pung or Kong before saying chow for {self.game.lastDiscard.name()}')
            return deferLater(Internal.reactor, delayStep, self.__delayAnswer, result, delay, delayStep)
        return succeed(result)

    def thatWasMe(self, player:PlayingPlayer) ->bool:
        """return True if player == myself"""
        if not self.game:
            return False
        return player == self.game.myself

    @staticmethod
    def __jellyMessage(value:str) ->str:
        """the Message classes are not pb.copyable, convert them into their names"""
        return Message.OK.name if value is None else Message.jelly(value, value)

    def remote_move(self, playerName:Optional[str], command:Union[Message, str],
		*unusedArgs:Any, **kwargs:Any) ->Deferred:
        """the server sends us info or a question and always wants us to answer"""
        if Internal.scene and not isAlive(Internal.scene):
            return fail()
        if self.game and playerName:
            player = cast(PlayingPlayer, self.game.playerByName(playerName))
        elif playerName:
            player = PlayingPlayer(None, playerName)
        else:
            player = None
        move = Move(player, command, kwargs)
        if Debug.traffic:
            if self.isHumanClient():
                if self.game:
                    self.game.debug(f'got Move: {move!r}')
                else:
                    logDebug(f'got Move: {move!r}')
        game = self.game
        if game:
            if move.token:
                game_token = game.point.token(game)
                if move.token != game_token:
                    logException(
                        f'wrong token: {move.token}, we have {game_token}')
        with Duration(f'{move!r}:'):
            try:
                result = self.exec_move(move).addCallback(self.__jellyMessage)
            except (Exception, Failure) as exc:  # pylint: disable=broad-exception-caught
                return fail(exc)
            return result

    def exec_move(self, move:Move) ->Deferred:
        """mirror the move of a player as told by the game server"""
        message = move.message
        if message.needsGame and not self.game:  # type: ignore
            # server already disconnected, see
            # HumanClient.remote_ServerDisconnects
            return succeed(Message.OK)
        action = message.notifyAction if move.notifying else message.clientAction  # type: ignore
        game = self.game
        if game:
            game.moves.append(move)
        answer = action(self, move)
        if not isinstance(answer, Deferred):
            answer = succeed(answer)
        if game:
            if not move.notifying and move.player and not move.player.scoreMatchesServer(move.score):
                game.close()
# This is an example how to find games where specific situations arise. We prefer games where this
# happens very early for easier reproduction. So set number of rounds to 1 in the ruleset before doing this.
# This example looks for a situation where the single human player may call Chow but one of the
# robot players calls Pung. See https://bugs.kde.org/show_bug.cgi?id=318981
#            if game.nextPlayer() == game.myself:
# I am next
#                if message == Message.Pung and move.notifying:
# somebody claimed a pung
#                    if move.player != game.myself:
# it was not me
#                        if game.point.handCount == 0 and len(game.moves) < 30:
# early on in the game
#                            game.myself.computeSayable(move, [Message.Chow])
#                            if game.myself.sayable[Message.Chow]:
# I may say Chow
#                                logDebug('FOUND EXAMPLE FOR %s IN %s' % (game.myself,
#                                       game.point.prompt(game, withMoveCount=True)))

        if message == Message.Discard:
            # do not block here, we want to get the clientDialog
            # before the animated tile reaches its end position
            animate()
            return answer
        if message == Message.AskForClaims:
            # no need to start an animation. If we did the below standard clause, this is what
            # could happen:
            # 1. user says Chow
            # 2. SelectChow dialog pops up
            # 3. previous animation ends, making animate() callback with current answer
            # 4. but this answer is Chow, without a selected Chow. This is
            # wrongly sent to server
            return answer
        # return answer only after animation ends. Put answer into
        # the Deferred returned by animate().
        return animate().addCallback(lambda x: answer).addErrback(logFailure)

    def claimed(self, move:Move) ->Optional[Deferred]:
        """somebody claimed a discarded tile"""
        assert self.game
        if Internal.scene:
            calledTileItem = cast('PlayingScene', Internal.scene).discardBoard.claimDiscard()
            calledTile = calledTileItem.tile
        else:
            calledTileItem = None
            assert self.game.lastDiscard
            calledTile = self.game.lastDiscard
        self.game.lastDiscard = None
        assert calledTile
        self.game.discardedTiles[calledTile.exposed] -= 1
        assert move.meld, f'move has no meld: {move!r}'
        assert calledTile in move.meld, f'{calledTile} {move.meld}'
        hadTiles = move.meld.without(calledTile)
        assert move.player
        if not self.thatWasMe(move.player) and not self.game.playOpen:
            move.player.showConcealedTiles(hadTiles)
        move.player.lastTile = calledTile.exposed
        move.player.lastSource = TileSource.LivingWallDiscard
        move.exposedMeld = move.player.exposeMeld(
            hadTiles, # type:ignore[arg-type]
            calledTile=calledTileItem or calledTile) # type:ignore[arg-type]

        if self.thatWasMe(move.player):
            if move.message != Message.Kong:
                # we will get a replacement tile first
                return self.myAction(move)
        elif self.game.prevActivePlayer == self.game.myself and self.connection:
            # even here we ask: if our discard is claimed we need time
            # to notice - think 3 robots or network timing differences
            return self.ask(move, [Message.OK])
        return None

    def myAction(self, move:Move) ->Deferred:
        """ask myself what I want to do after picking or claiming a tile"""
        # only when all animations ended, our handboard gets focus. Otherwise
        # we would see a blue focusRect in the handboard even when a tile
        # ist still moving from the discardboard to the handboard.
        assert move.player
        animateAndDo(move.player.getsFocus)
        possibleAnswers = [Message.Discard, Message.Kong, Message.MahJongg]
        if not move.player.discarded:
            possibleAnswers.append(Message.OriginalCall)
        return self.ask(move, list(cast('ClientMessage', x) for x in possibleAnswers))

    def declared(self, move:Move) ->None:
        """somebody declared something.
        By declaring we mean exposing a meld, using only tiles from the hand.
        For now we only support Kong: in Classical Chinese it makes no sense
        to declare a Pung."""
        assert move.message == Message.Kong
        assert self.game
        assert move.player
        if not self.thatWasMe(move.player) and not self.game.playOpen:
            move.player.showConcealedTiles(move.meld)
        move.exposedMeld = move.player.exposeMeld(move.meld)
        if not self.thatWasMe(move.player):
            self.ask(move, [Message.OK])

    def __str__(self) ->str:
        assert self.name
        return self.name



# -*- coding: utf-8 -*-

"""
Copyright (C) 2009-2016 Wolfgang Rohdewald <wolfgang@rohdewald.de>

SPDX-License-Identifier: GPL-2.0-only


The DBPasswordChecker is based on an example from the book
Twisted Network Programming Essentials by Abe Fettig, 2006
O'Reilly Media, Inc., ISBN 0-596-10032-9
"""

import sys
import random
import traceback
from itertools import chain
import sqlite3

from typing import TYPE_CHECKING, Any, Optional, List, Callable, Tuple, Dict, cast, Union

from common import Debug, Internal, ReprMixin
from wind import Wind
from tilesource import TileSource
from util import Duration
from log import logDebug, logError
from mi18n import i18nE, i18n, i18ncE
from tile import Tile, Piece, TileTuple, elements, Meld, MeldList
from query import Query
from client import Client, Table
from wall import WallEmpty
from sound import Voice
from servercommon import srvError
from user import User
from game import PlayingGame
from message import Message, ChatMessage, ServerMessage
from deferredutil import DeferredBlock
from point import PointRange

if TYPE_CHECKING:
    from deferredutil import Request
    from player import Player, PlayingPlayer
    from rule import Ruleset
    import datetime
    from server import MJServer


if sys.platform != 'win32':
    import resource


class ServerGame(PlayingGame):

    """the central game instance on the server"""

    def __init__(self, names:List[Tuple[Wind, str]], ruleset:'Ruleset',
        gameid:Optional[int]=None, wantedGame:Optional[str]=None,
        client:Optional[Client]=None, playOpen:bool=False, autoPlay:bool=False) ->None:
        PlayingGame.__init__(
            self,
            names,
            ruleset,
            gameid,
            wantedGame,
            client,
            playOpen,
            autoPlay)
        self.shouldSave = True

    def throwDices(self) ->None:
        """set random living and kongBox
        sets divideAt: an index for the wall break"""
        assert self.wall
        self.wall.tiles.sort()
        self.randomGenerator.shuffle(self.wall.tiles)
        super().throwDices()

    def initHand(self) ->None:
        """Happens only on server: every player gets 13 tiles (including east)"""
        assert self.wall
        self.throwDices()
        self.wall.divide()
        for player in self.players:
            player.clearHand()
            # 13 tiles at least, with names as given by wall
            # compensate boni
            while len(player.concealedTiles) != self.ruleset.dealtTiles:
                player.addConcealedTiles(self.wall.deal())
        super().initHand()

    def _concealedTileName(self, tileName:Tile) ->Tile:
        """The server game instance knows everything but has no myself"""
        player = self.activePlayer
        if tileName not in player.concealedTiles:
            raise ValueError(f'I am the server Game instance. Player {player.name} is told '
                             f'to show discard of tile {tileName} but does not have it, he has {player.concealedTiles}')
        return tileName

class ServerTable(Table, ReprMixin):

    """a table on the game server"""
    # pylint: disable=too-many-arguments

    def __init__(self, server:'MJServer', owner:Optional[User], ruleset:str,
                 suspendedAt:Optional[str],
                 playOpen:bool, autoPlay:bool, wantedGame:str, tableId:Optional[int]=None) ->None:
        if tableId is None:
            tableId = server.generateTableId()
        Table.__init__(
            self,
            tableId,
            ruleset,
            suspendedAt,
            False,
            playOpen,
            autoPlay,
            wantedGame)
        self.server = server
        self.owner = owner
        self.users = [owner] if owner else []
        self.remotes:Dict['PlayingPlayer', Union[User, Client]] = {}   # maps client connections to users
        self.game:Optional[ServerGame] = None
        self.client:Optional[Client] = None
        server.tables[self.tableid] = self
        if Debug.table:
            logDebug(f'new table {self}')

    def hasName(self, name:str) ->bool:
        """return True if one of the players in the game is named 'name'"""
        assert self.game
        return bool(self.game) and any(x.name == name for x in self.game.players)

    def asSimpleList(self, withFullRuleset:bool=False) ->List[Any]:
        """return the table attributes to be sent to the client"""
        game = self.game
        onlineNames = [x.name for x in self.users]
        if self.suspendedAt:
            assert game
            names = tuple(x.name for x in game.players)
        else:
            names = tuple(x.name for x in self.users)
        online = tuple(bool(x in onlineNames) for x in names)
        if game:
            endValues = game.point.handCount, {x.wind.char: x.balance for x in game.players}
        else:
            endValues = None
        return list([
            self.tableid,
            self.ruleset.toList() if withFullRuleset else self.ruleset.hash,
            game.gameid if game else None, self.suspendedAt, self.running,
            self.playOpen, self.autoPlay, self.wantedGame, names, online, endValues])

    def maxSeats(self) ->int:
        """for a new game: 4. For a suspended game: The
        number of humans before suspending"""
        result = 4
        if self.suspendedAt:
            assert self.game
            result -= sum(x.name.startswith('Robot ')
                          for x in self.game.players)
        return result

    def sendChatMessage(self, chatLine:ChatMessage) ->None:
        """sends a chat messages to all clients"""
        assert self.game
        if Debug.chat:
            logDebug(f'server sends chat msg {chatLine}')
        if self.suspendedAt and self.game:
            chatters:List['User'] = []
            for player in self.game.players:
                chatters.extend(
                    x for x in self.server.srvUsers if x.name == player.name)
        else:
            chatters = self.users
        for other in chatters:
            self.server.callRemote(other, 'chat', chatLine.asList())

    def addUser(self, user:User) ->None:
        """add user to this table"""
        if user.name in (x.name for x in self.users):
            raise srvError(i18nE('You already joined this table'))
        if len(self.users) == self.maxSeats():
            raise srvError(i18nE('All seats are already taken'))
        self.users.append(user)
        if Debug.table:
            logDebug(f'{user.name} seated on table {self}')
        self.sendChatMessage(ChatMessage(self.tableid, user.name,
                                         i18nE('takes a seat'), isStatusMessage=True))

    def delUser(self, user:User) ->None:
        """remove user from this table"""
        if user in self.users:
            self.running = False
            self.users.remove(user)
            self.sendChatMessage(ChatMessage(self.tableid, user.name,
                                             i18nE('leaves the table'), isStatusMessage=True))
            if user is self.owner:
                # silently pass ownership
                if self.users:
                    self.owner = self.users[0]
                    if Debug.table:
                        logDebug(f'{user.name} leaves table {int(self.tableid)}, {self.owner} is the new owner')
                else:
                    if Debug.table:
                        logDebug(f'{user.name} leaves table {int(self.tableid)}, table is now empty')
            else:
                if Debug.table:
                    logDebug(f'{user.name} leaves table {int(self.tableid)}, {self.owner} stays owner')

    def __str__(self)->str:
        """for debugging output"""
        if self.users:
            onlineNames = [x.name + ('(Owner)' if self.owner and x == self.owner.name else '')
                           for x in self.users]
        else:
            onlineNames = ['no users yet']
        offlineString = ''
        if self.game:
            offlineNames = [x.name for x in self.game.players if x.name not in onlineNames
                            and not x.name.startswith('Robot')]
            if offlineNames:
                offlineString = ' offline:' + ','.join(offlineNames)
        return f"{int(self.tableid)}({','.join(onlineNames)}{offlineString})"

    def calcGameId(self)->int:
        """based upon the max gameids we got from the clients, propose
        a new one, we want to use the same gameid in all data bases"""
        serverMaxGameId = Query('select max(id) from game').records[0][0]
        serverMaxGameId = int(serverMaxGameId) if serverMaxGameId else 0
        gameIds = [serverMaxGameId]
        gameIds.extend(x.maxGameId for x in self.users)
        return max(gameIds) + 1

    def __prepareNewGame(self)->ServerGame:
        """return a new game object"""
        names = [x.name for x in self.users]
        # the server and all databases save the english name but we
        # want to make sure a translation exists for the client GUI
        robotNames = [
            i18ncE(
                'kajongg, name of robot player, to be translated',
                'Robot 1'),
            i18ncE(
                'kajongg, name of robot player, to be translated',
                'Robot 2'),
            i18ncE('kajongg, name of robot player, to be translated', 'Robot 3')]
        while len(names) < 4:
            names.append(robotNames[3 - len(names)])
        names = [tuple([Wind.all4[idx], name]) for idx, name in enumerate(names)]
        self.client = Client()
                             # Game has a weakref to client, so we must keep
                             # it!
        return ServerGame(names, self.ruleset, client=self.client,
                          playOpen=self.playOpen, autoPlay=self.autoPlay, wantedGame=self.wantedGame)

    def userForPlayer(self, player:'PlayingPlayer') ->Optional[User]:
        """finds the table user corresponding to player"""
        for result in self.users:
            if result.name == player.name:
                return result
        return None

    def __connectPlayers(self) ->None:
        """connects client instances with the game players"""
        game = self.game
        assert game
        for player in game.players:
            remote:Union[User, Client, None] = self.userForPlayer(player)
            if not remote:
                # we found a robot player, its client runs in this server
                # process
                remote = Client(player.name)
                remote.table = self
            self.remotes[player] = remote

    def __checkDbIdents(self) ->None:
        """for 4 players, we have up to 4 data bases:
        more than one player might use the same data base.
        However the server always needs to use its own data base.
        If a data base is used by more than one client, only one of
        them should update. Here we set shouldSave for all players,
        while the server always saves"""
        serverIdent = Internal.db.identifier
        dbIdents = set()
        game = self.game
        assert game
        for player in game.players:
            player.shouldSave = False
            _ = self.remotes[player]
            if isinstance(_, User):
                dbIdent = _.dbIdent
                assert dbIdent != serverIdent, \
                    f'client and server try to use the same database:{Internal.db.path}'
                player.shouldSave = dbIdent not in dbIdents
                dbIdents.add(dbIdent)

    def readyForGameStart(self, user:User) ->None:
        """the table initiator told us he wants to start the game"""
        assert self.owner
        if len(self.users) < self.maxSeats() and self.owner != user:
            raise srvError(i18nE(
                               'Only the initiator %1 can start this game, you are %2'),
                           self.owner.name, user.name)
        if self.suspendedAt:
            self.__connectPlayers()
            self.__checkDbIdents()
            self.initGame()
        else:
            self.game = self.__prepareNewGame()
            self.__connectPlayers()
            self.__checkDbIdents()
            self.proposeGameId(self.calcGameId())
        # TODO: remove table for all other srvUsers out of sight

    def proposeGameId(self, gameid:int) ->None:
        """server proposes an id to the clients ands waits for answers"""
        counter = 0
        while True:
            try:
                Query('insert into game(id,seed) values(?,?)',
                          (gameid, 'proposed'), mayFail=True, failSilent=True)
                break
            except sqlite3.IntegrityError:
                if counter > 100:
                    raise
                counter += 1
                gameid += random.randrange(1, 10)
        block = DeferredBlock(self, where='proposeGameId')
        assert self.game
        for player in self.game.players:
            if player.shouldSave and isinstance(self.remotes[player], User):
                # do not ask robot players, they use the server data base
                block.tellPlayer(player, Message.ProposeGameId, gameid=gameid)
        block.callback(self.collectGameIdAnswers, gameid)

    def collectGameIdAnswers(self, requests:List['Request'], gameid:int) ->None:
        """clients answered if the proposed game id is free"""
        if requests:
            # when errors happen, there might be no requests left
            assert self.game
            for msg in requests:
                if msg.answer == Message.NO:
                    self.proposeGameId(gameid + 1)
                    return
                if msg.answer != Message.OK:
                    raise srvError(
                        'collectGameIdAnswers got neither NO nor OK')
            self.game.gameid = gameid
            self.initGame()

    def initGame(self) ->None:
        """ask clients if they are ready to start"""
        game = self.game
        assert game
        game.saveStartTime()
        block = DeferredBlock(self, where='initGame')
        for player in game.players:
            block.tellPlayer(
                player, Message.ReadyForGameStart, tableid=self.tableid,
                gameid=game.gameid, shouldSave=player.shouldSave,
                wantedGame=str(PointRange(game, game.first_point, game.last_point)), players=game.players)
        block.callback(self.startGame)

    def startGame(self, requests:List['Request']) ->None:
        """if all players said ready, start the game"""
        assert self.game
        for user in self.users:
            userRequests = [x for x in requests if x.user == user]
            if not userRequests or userRequests[0].answer == Message.NoGameStart:
                if Debug.table:
                    if not userRequests:
                        logDebug(
                            f'Server.startGame: found no request for user {user.name}')
                    else:
                        logDebug(
                            f'Server.startGame: {user.name} said NoGameStart')
                self.game = None
                return
        if Debug.table:
            logDebug(f'Game starts on table {self}')
        elementIter = iter(elements.all(self.game.ruleset))
        assert self.game.wall
        self.game.wall.tiles = []
        for _ in range(self.game.fullWallSize):
            self.game.wall.tiles.append(Piece(next(elementIter).concealed))
        assert isinstance(self.game, ServerGame), self.game
        self.running = True
        self.__adaptOtherTables()
        self.sendVoiceIds()

    def __adaptOtherTables(self) ->None:
        """if the players on this table also reserved seats on other tables, clear them
        make running table invisible for other users"""
        for user in self.users:
            for tableid in self.server.tablesWith(user):
                if tableid != self.tableid:
                    self.server.leaveTable(user, tableid, f'now sits on table {self.tableid}')
        foreigners = [x for x in self.server.srvUsers if x not in self.users]
        if foreigners:
            if Debug.table:
                logDebug(
                    f"make running table {self} invisible for {','.join(str(x) for x in foreigners)}")
            for srvUser in foreigners:
                self.server.callRemote(
                    srvUser,
                    'tableRemoved',
                    self.tableid,
                    '')

    def sendVoiceIds(self) ->None:
        """tell each player what voice ids the others have. By now the client has a Game instance!"""
        assert self.game
        humanPlayers = [
            x for x in self.game.players if isinstance(self.remotes[x], User)]
        if len(humanPlayers) < 2 or not any(self.remotes[x].voiceId for x in humanPlayers):
            # no need to pass around voice data
            self.assignVoices()
            return
        block = DeferredBlock(self, where='sendVoiceIds')
        for player in humanPlayers:
            remote = self.remotes[player]
            if remote.voiceId:
                # send it to other human players:
                others = [x for x in humanPlayers if x != player]
                if Debug.sound:
                    logDebug(f'telling other human players that {player.name} has voiceId {remote.voiceId}')
                block.tell(
                    player,
                    others,
                    Message.VoiceId,
                    source=remote.voiceId)
        block.callback(self.collectVoiceData)

    def collectVoiceData(self, requests:List['Request']) ->None:
        """collect voices of other players"""
        if not self.running:
            return
        assert self.game
        block = DeferredBlock(self, where='collectVoiceData')
        voiceDataRequests = []
        for request in requests:
            if request.answer == Message.ClientWantsVoiceData:
                # another human player requests sounds for voiceId
                assert request.args
                voiceId = request.args[0]
                voiceFor = [x for x in self.game.players if isinstance(self.remotes[x], User)
                            and self.remotes[x].voiceId == voiceId][0]
                voiceFor.voice = Voice(voiceId)
                if Debug.sound:
                    logDebug(
                        f'client {request.user.name} wants voice data {request.args[0]} for {voiceFor}')
                voiceDataRequests.append((request.user, voiceFor))
                if not voiceFor.voice.oggFiles():
                    # the server does not have it, ask the client with that
                    # voice
                    block.tell(
                        voiceFor,
                        [voiceFor],
                        Message.ServerWantsVoiceData)
        block.callback(self.sendVoiceData, voiceDataRequests)

    def sendVoiceData(self, requests:List['Request'], voiceDataRequests:List[Tuple[User, 'PlayingPlayer']]) ->None:
        """sends voice sounds to other human players"""
        self.processAnswers(requests)
        block = DeferredBlock(self, where='sendVoiceData')
        for voiceDataRequester, voiceFor in voiceDataRequests:
            # this player requested sounds for voiceFor
            voice = voiceFor.voice
            assert voice
            content = voice.archiveContent
            if content:
                if Debug.sound:
                    logDebug(
                        f'server got voice data {voiceFor.voice} for {voiceFor.name} from client')
                block.tell(
                    voiceFor,
                    [voiceDataRequester],
                    Message.VoiceData,
                    md5sum=voice.md5sum,
                    source=content)
            elif Debug.sound:
                logDebug(f'server got empty voice data {voice} for {voiceFor.name} from client')
        block.callback(self.assignVoices)

    def assignVoices(self, unusedResults:Optional[List['Request']]=None) ->None:
        """now all human players have all voice data needed"""
        assert self.game
        humanPlayers = [
            x for x in self.game.players if isinstance(self.remotes[x], User)]
        block = DeferredBlock(self, where='assignVoices')
        block.tell(None, humanPlayers, Message.AssignVoices)
        block.callback(self.startHand)

    def __tilePicked(self, unusedResults:List['Request'], deadEnd:bool=False) ->None:
        """from callback"""
        self.pickTile(deadEnd)

    def pickTile(self, deadEnd:bool=False) ->None:
        """the active player gets a tile from wall. Tell all clients."""
        if not self.running:
            return
        assert self.game
        player = self.game.activePlayer
        try:
            tile = player.pickedTile(deadEnd)
        except WallEmpty:
            self.endHand()
        else:
            self.game.lastDiscard = None
            block = DeferredBlock(self, where='pickTile')
            block.tellPlayer(
                player,
                Message.PickedTile,
                tile=tile,
                deadEnd=deadEnd)
            showTile = tile if tile.isBonus or self.game.playOpen else Tile.unknown
            block.tellOthers(
                player,
                Message.PickedTile,
                tile=showTile,
                deadEnd=deadEnd)
            block.callback(self.moved)

    def __pickKongReplacement(self, requests:List['Request']) ->None:
        """the active player gets a tile from the dead end. Tell all clients."""
        requests = self.prioritize(requests)
        if requests and requests[0].answer == Message.MahJongg:
            # somebody robs my kong
            Message.MahJongg.serverAction(self, requests[0])  # type:ignore[call-arg, arg-type]
        else:
            self.pickTile(deadEnd=True)

    def clientDiscarded(self, msg:'Request') ->None:
        """client told us he discarded a tile. Check for consistency and tell others."""
        if not self.running:
            return
        assert self.game
        player = msg.player
        assert player == self.game.activePlayer
        assert msg.args
        tile = Tile(msg.args[0])
        if tile not in player.concealedTiles:
            self.abort(
                f'player {player} discarded {tile} but does not have it. He has {player.concealedTiles}')
            return
        dangerousText = self.game.dangerousFor(player, tile)
        mustPlayDangerous = player.mustPlayDangerous()
        violates = player.violatesOriginalCall(tile)
        self.game.hasDiscarded(player, tile)
        block = DeferredBlock(self, where='clientDiscarded')
        block.tellAll(player, Message.Discard, tile=tile)
        block.callback(self._clientDiscarded2, msg, dangerousText, mustPlayDangerous, violates)

    def _clientDiscarded2(self, unusedResults:List['Request'], msg:'Request',
        dangerousText:str, mustPlayDangerous:bool, violates:bool) ->None:
        """client told us he discarded a tile. Continue, check for violating original call"""
        block = DeferredBlock(self, where='_clientDiscarded2')
        player = msg.player
        assert player
        if violates:
            if Debug.originalCall:
                assert msg.args
                tile = Tile(msg.args[0])
                logDebug(f'{player} just violated OC with {tile}')
            player.mayWin = False
            block.tellAll(player, Message.ViolatesOriginalCall)
        block.callback(self._clientDiscarded3, msg, dangerousText, mustPlayDangerous)

    def _clientDiscarded3(self, unusedResults:List['Request'], msg:'Request',
        dangerousText:str, mustPlayDangerous:bool) ->None:
        """client told us he discarded a tile. Continue, check for calling"""
        assert self.game
        block = DeferredBlock(self, where='_clientDiscarded3')
        player = msg.player
        assert player
        if self.game.ruleset.mustDeclareCallingHand and not player.isCalling:
            if player.hand.callingHands:
                player.isCalling = True
                block.tellAll(player, Message.Calling)
        block.callback(self._clientDiscarded4, msg, dangerousText, mustPlayDangerous)

    def _clientDiscarded4(self, unusedResults:List['Request'], msg:'Request',
        dangerousText:str, mustPlayDangerous:bool) ->None:
        """client told us he discarded a tile. Continue, check for dangerous game"""
        block = DeferredBlock(self, where='_clientDiscarded4')
        player = msg.player
        assert player
        assert msg.args
        if dangerousText:
            if mustPlayDangerous and not player.lastSource.isDiscarded:
                if Debug.dangerousGame:
                    tile = Tile(msg.args[0])
                    logDebug(f"{player} claims no choice. Discarded {tile}, "
                             f"keeping {player.concealedTiles}. {' / '.join(dangerousText)}")
                player.claimedNoChoice = True
                block.tellAll(
                    player,
                    Message.NoChoice,
                    tiles=TileTuple(player.concealedTiles))
            else:
                player.playedDangerous = True
                if Debug.dangerousGame:
                    tile = Tile(msg.args[0])
                    logDebug(f"{player} played dangerous. Discarded {tile}, "
                             f"keeping {player.concealedTiles}. {' / '.join(dangerousText)}")
                block.tellAll(
                    player,
                    Message.DangerousGame,
                    tiles=TileTuple(player.concealedTiles))
        if msg.answer == Message.OriginalCall:
            player.isCalling = True
            block.callback(self.clientMadeOriginalCall, msg)
        else:
            block.callback(self._askForClaims, msg)

    def clientMadeOriginalCall(self, unusedResults:List['Request'], msg:'Request') ->None:
        """first tell everybody about original call
        and then treat the implicit discard"""
        assert msg.player
        msg.player.originalCall = True
        if Debug.originalCall:
            logDebug(f'server.clientMadeOriginalCall: {msg.player}')
        block = DeferredBlock(self, where='clientMadeOriginalCall')
        block.tellAll(msg.player, Message.OriginalCall)
        block.callback(self._askForClaims, msg)

    def startHand(self, unusedResults:Optional[List['Request']]=None) ->None:
        """all players are ready to start a hand, so do it"""
        if self.running:
            assert self.game
            self.game.prepareHand()
            self.game.initHand()
            block = self.tellAll(None, Message.InitHand,
                                 divideAt=self.game.divideAt)
            block.callback(self.divided)

    def divided(self, unusedResults:List['Request']) ->None:
        """the wall is now divided for all clients"""
        if not self.running:
            return
        assert self.game
        block = DeferredBlock(self, where='divided')
        for clientPlayer in self.game.players:
            for player in self.game.players:
                if player == clientPlayer or self.game.playOpen:
                    tiles = player.concealedTiles
                else:
                    tiles = TileTuple(Tile.unknown * self.game.ruleset.dealtTiles)  # type: ignore
                block.tell(player, [clientPlayer], Message.SetConcealedTiles,
                           tiles=TileTuple(chain(tiles, player.bonusTiles)))
        block.callback(self.dealt)

    def handEnded(self, unusedResults:List['Request']) ->None:
        """from callback"""
        self.endHand()

    def endHand(self) ->None:
        """hand is over, show all concealed tiles to all players"""
        if not self.running:
            return
        assert self.game
        if self.game.playOpen:
            self.saveHand()
        else:
            block = DeferredBlock(self, where='endHand')
            for player in self.game.players:
                # there might be no winner, winner.others() would be wrong
                if player != self.game.winner:
                    # the winner tiles are already shown in claimMahJongg
                    block.tellOthers(
                        player, Message.ShowConcealedTiles, show=True,
                        tiles=TileTuple(player.concealedTiles))
            block.callback(self.saveHand)

    def saveHand(self, unusedResults:Optional[List['Request']]=None) ->None:
        """save the hand to the database and proceed to next hand"""
        if not self.running:
            return
        assert self.game
        self.tellAll(None, Message.SaveHand, self.nextHand)
        self.game.saveHand()

    def nextHand(self, unusedResults:List['Request']) ->None:
        """next hand: maybe rotate"""
        if not self.running:
            return
        assert self.game
        DeferredBlock.garbageCollection()
        for block in DeferredBlock.blocks:
            if block.table == self:
                logError(
                    f'request left from previous hand: {block.outstandingStr()}')
        token = self.game.point.prompt(self.game,
            withAI=False)  # we need to send the old token until the
                                   # clients started the new hand
        mustRotateWinds = self.game.maybeRotateWinds()
        if self.game.finished():
            self.server.removeTable(
                self,
                'gameOver',
                i18nE('Game <numid>%1</numid> is over!'),
                self.game.seed)
            if Debug.process and sys.platform != 'win32':
                logDebug(
                    f'MEM:{resource.getrusage(resource.RUSAGE_SELF).ru_maxrss}')  # pylint:disable=possibly-used-before-assignment
            return
        self.game.sortPlayers()
        self.tellAll(None, Message.ReadyForHandStart, self.startHand,
                     players=self.game.players, mustRotateWinds=mustRotateWinds, token=token)

    def abort(self, message:str, *args:Any) ->None:
        """abort the table. Reason: message/args"""
        self.server.removeTable(self, 'abort', message, *args)

    def claimTile(self, player:'PlayingPlayer', claim:'ServerMessage',
        meld:Meld, nextMessage:Message) ->None:
        """a player claims a tile for pung, kong or chow.
        meld contains the claimed tile, concealed"""
        if not self.running:
            return
        assert self.game
        assert meld.isConcealed
        lastDiscard = self.game.lastDiscard
        # if we rob a tile, self.game.lastDiscard has already been set to the
        # robbed tile
        discardingPlayer = self.game.activePlayer
        assert lastDiscard
        hasTiles = meld.without(lastDiscard)
        if len(meld) != 4 and not (meld.isPair or meld.isPungKong or meld.isChow):
            msg = i18nE('%1 wrongly said %2 for meld %3')
            self.abort(msg, player.name, claim.name, str(meld))
            return
        if not player.hasConcealedTiles(hasTiles):
            msg = i18nE(
                '%1 wrongly said %2: claims to have concealed tiles %3 but only has %4')
            self.abort(
                msg,
                player.name,
                claim.name,
                ' '.join(hasTiles),
                str(player.concealedTiles))
            return
        # update our internal state before we listen to the clients again
        self.game.discardedTiles[lastDiscard.exposed] -= 1
        self.game.activePlayer = player
        assert lastDiscard
        player.exposeMeld(hasTiles, lastDiscard)  # type:ignore[arg-type]
        if lastDiscard:
            player.lastTile = lastDiscard.exposed
            player.lastSource = TileSource.LivingWallDiscard
        self.game.lastDiscard = None
        block = DeferredBlock(self, where='claimTile')
        if (nextMessage != Message.Kong
                and self.game.dangerousFor(discardingPlayer, lastDiscard)
                and discardingPlayer.playedDangerous):
            player.usedDangerousFrom = discardingPlayer
            if Debug.dangerousGame:
                logDebug(f'{player} claims dangerous tile {lastDiscard} discarded by {discardingPlayer}')
            block.tellAll(
                player,
                Message.UsedDangerousFrom,
                source=discardingPlayer.name)
        block.tellAll(player, nextMessage, meld=meld)
        if claim == Message.Kong:
            block.callback(self.__pickKongReplacement)
        else:
            block.callback(self.moved)

    def declareKong(self, player:'PlayingPlayer', kongMeld:Meld) ->None:
        """player declares a Kong"""
        if not player.hasConcealedTiles(kongMeld) and kongMeld[0].exposed.pung not in player.exposedMelds:
            msg = i18nE('declareKong:%1 wrongly said Kong for meld %2')
            args = (player.name, str(kongMeld))
            logDebug(i18n(msg, *args))
            logDebug(f"declareKong:concealedTiles:{player.concealedTiles}")
            logDebug(f"declareKong:concealedMelds:{player.concealedMelds}")
            logDebug(f"declareKong:exposedMelds:{player.exposedMelds}")
            self.abort(msg, *args)
            return
        player.exposeMeld(kongMeld)
        self.tellAll(
            player,
            Message.DeclaredKong,
            self.__pickKongReplacement,
            meld=kongMeld)

    def claimMahJongg(self, msg:'Request') ->None:
        """a player claims mah jongg. Check this and
        if correct, tell all. Otherwise abort game, kajongg client is faulty"""
        if not self.running:
            return
        assert self.game
        player = msg.player
        assert player
        assert msg.args
        concealedMelds = MeldList(msg.args[0])
        withDiscard:Optional[Tile] = Tile(msg.args[1]) if msg.args[1] else None
        lastMeld = Meld(msg.args[2])
        if self.game.ruleset.mustDeclareCallingHand:
            assert player.isCalling, (f'{self.game.point} {player.hand} {player} says MJ but never claimed: '
                                     f'concmelds:{concealedMelds} withdiscard:{withDiscard} lastmeld:{lastMeld}')
        discardingPlayer = self.game.activePlayer
        lastMove = next(self.game.lastMoves(withoutNotifications=True))
        robbedTheKong = lastMove.message == Message.DeclaredKong
        if robbedTheKong:
            player.robsTile()
            assert lastMove.player
            _ = lastMove.meld[0].concealed
            lastMove.player.getsRobbed(_)
            withDiscard = _
        msgArgs = player.showConcealedMelds(concealedMelds, withDiscard)
        if msgArgs:
            self.abort(*msgArgs)
            return
        player.declaredMahJongg(
            concealedMelds,
            withDiscard,
            player.lastTile,
            lastMeld)
        if not player.hand.won:
            message = i18nE('%1 claiming MahJongg: This is not a winning hand: %2')
            self.abort(message, player.name, player.hand.string)
            return
        block = DeferredBlock(self, where='claimMahJongg')
        if robbedTheKong:
            block.tellAll(player, Message.RobbedTheKong, tile=withDiscard)
        if (player.lastSource is TileSource.LivingWallDiscard
                and self.game.dangerousFor(discardingPlayer, player.lastTile)
                and discardingPlayer.playedDangerous):
            player.usedDangerousFrom = discardingPlayer
            if Debug.dangerousGame:
                logDebug(f'{player} wins with dangerous tile {self.game.lastDiscard} from {discardingPlayer}')
            block.tellAll(
                player,
                Message.UsedDangerousFrom,
                source=discardingPlayer.name)
        block.tellAll(
            player, Message.MahJongg, melds=concealedMelds, lastTile=player.lastTile,
            lastMeld=lastMeld, withDiscardTile=withDiscard)
        block.callback(self.handEnded)

    def dealt(self, unusedResults: List['Request']) ->None:
        """all tiles are dealt, ask east to discard a tile"""
        if self.running:
            assert self.game
            self.tellAll(
                self.game.activePlayer,
                Message.ActivePlayer,
                self.__tilePicked)

    def nextTurn(self) ->None:
        """the next player becomes active"""
        if self.running:
            # the player might just have disconnected
            assert self.game
            self.game.nextTurn()
            self.tellAll(
                self.game.activePlayer,
                Message.ActivePlayer,
                self.__tilePicked)

    def prioritize(self, requests: List['Request']) ->List['Request']:
        """return only requests we want to execute"""
        if not self.running:
            return []
        answers = [
            x for x in requests if x.answer not in [
                Message.NoClaim,
                Message.OK,
                None]]
        if len(answers) > 1:
            claims = [
                Message.MahJongg,
                Message.Kong,
                Message.Pung,
                Message.Chow]
            for claim in claims:
                if claim in [x.answer for x in answers]:
                    # ignore claims with lower priority:
                    answers = [
                        x for x in answers if x.answer == claim or x.answer not in claims]
                    break
        mjAnswers = [x for x in answers if x.answer == Message.MahJongg]
        if len(mjAnswers) > 1:
            assert self.game
            mjPlayers = [x.player for x in mjAnswers]
            nextPlayer = self.game.nextPlayer()
            while nextPlayer not in mjPlayers:
                nextPlayer = self.game.nextPlayer(nextPlayer)
            answers = [
                x for x in answers if x.player == nextPlayer or x.answer != Message.MahJongg]
        return answers

    def _askForClaims(self, unusedRequests:List['Request'], unusedMsg:'ServerMessage') ->None:
        """ask all players if they want to claim"""
        if self.running:
            assert self.game
            self.tellOthers(
                self.game.activePlayer,
                Message.AskForClaims,
                self.moved)

    def processAnswers(self, requests: List['Request']) ->List['Request']:
        """a player did something"""
        if not self.running:
            return []
        answers = self.prioritize(requests)
        if not answers:
            return []
        for answer in answers:
            msg = f'<-  {answer}'
            if Debug.traffic:
                logDebug(msg)
            with Duration(msg):
                assert answer.answer
                cast(ServerMessage, answer.answer).serverAction(self, answer)
        return answers

    def moved(self, requests: List['Request']) ->None:
        """a player did something"""
        if Debug.stack:
            stck = traceback.extract_stack()
            if len(stck) > 30:
                logDebug(f'stack size:{len(stck)}')
                logDebug(str(stck))
        answers = self.processAnswers(requests)
        if not answers:
            self.nextTurn()

    def tellAll(self, player:Optional['PlayingPlayer'], command:Message,
        callback:Optional[Callable[[List['Request']], None]]=None, **kwargs: Any) ->DeferredBlock:
        """tell something about player to all players"""
        block = DeferredBlock(self, where=f'tellAll: Player {player} command {command} kwargs {kwargs}')
        block.tellAll(player, command, **kwargs)
        block.callback(callback)
        return block

    def tellOthers(self, player:'PlayingPlayer', command:Message,
        callback:Optional[Callable[[List['Request']], None]]=None, **kwargs: Any) ->DeferredBlock:
        """tell something about player to all other players"""
        block = DeferredBlock(self, where='tellOthers')
        block.tellOthers(player, command, **kwargs)
        block.callback(callback)
        return block
